#include <time.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/timeb.h>
#include "logfile.h"

using namespace LogFile;

Log_File::Log_File(const string & file_name)
    : p_log_stream()
{
    rotate_log_files(file_name, NUMBER_OF_LOG_BACKUPS, 0);
    p_log_stream.open(file_name.c_str());
}

void
Log_File::write_line(const string & type, const string & line)
{
    char    ascii_now[80];
    char    ascii_ms[10];
    struct timeb    now;
    struct tm *     tm_now;


    // This seems to be the only way to get the milliseconds
    ftime(&now);
    tm_now  = localtime(&now.time);
    strftime(ascii_now, sizeof(ascii_now), "%e %h %Y %X", tm_now);
    sprintf(ascii_ms, "%03d", now.millitm);
    
    p_log_stream << ascii_now << "." << ascii_ms << " [" << type << "] "
                 << line << endl;
    p_log_stream.flush();
    return;
}

void
Log_File::error(const string & error_string)
{
    write_line("Err ", error_string);
    return;
}

void
Log_File::warning(const string & warning_string)
{
    write_line("Warn", warning_string);
    return;
}

void
Log_File::info(const string & info_string)
{
    write_line("Info", info_string);
    return;
}

Log_File *      LogFile::gl_log;

#if LOG_FILE_ENABLE_DEBUG
void
Log_File::debug(const string & info_string)
{
    write_line("Debug", info_string);
    return;
}
#endif

void
LogFile::open_log_file(const string & file_name)
{
    gl_log  = new Log_File(file_name);
}


void
LogFile::rotate_log_files(string base_name, int number_of_backups,
                          Log_Rotate_Listener * listener)
{
    char *  base_file;
    char *  pre_rotated_file;
    char *  post_rotated_file;
    char    extension[4];
    int     name_size;
    struct stat s;
    
    name_size   = base_name.length()+1;
    base_file   = new char[name_size];
    pre_rotated_file = new char[name_size];
    post_rotated_file = new char[name_size];
    strncpy(base_file, base_name.c_str(), name_size);
    
    for (int i = number_of_backups; i > 0; i--)
    {
        if (i == 1)
        {
            strncpy(pre_rotated_file, base_file, name_size);
        }
        else
        {
            sprintf(extension, "%03d", i-1);
            replace_extension(pre_rotated_file, base_file, extension,
                              name_size);
        }
        
        sprintf(extension, "%03d", i);
        replace_extension(post_rotated_file, base_file, extension, name_size);
        
        if (stat(pre_rotated_file, &s) == 0)
        {
            remove(post_rotated_file);
            rename(pre_rotated_file, post_rotated_file);
        }

        if (listener != 0)
        {
            Log_Rotate_Event e(i, number_of_backups);
            listener->rotate_progress(&e);
        }
    }

    delete post_rotated_file;
    delete pre_rotated_file;
    delete base_file;
}

void
LogFile::replace_extension(char * dest, const char * src,
                            const char * ext, int size)
{
    char * dest_ext;

    strncpy(dest, src, size);
    dest_ext = strrchr(dest, '.');
    if (strlen(dest_ext) == 4)
    {
        dest_ext++;
        strncpy(dest_ext, ext, 3);
    }
}


Log_Rotate_Listener::~Log_Rotate_Listener()
{
}

Log_Rotate_Event::Log_Rotate_Event(int completed_rotations,
                                   int total_rotations)
{
    p_completed_rotations   = completed_rotations;
    p_total_rotations       = total_rotations;
}

int
Log_Rotate_Event::get_completed_rotations() const
{
    return p_completed_rotations;
}

int Log_Rotate_Event::get_total_rotations() const
{
    return p_total_rotations;
}
