#include <iostream>
#include <strstream>
#include <stdlib.h>
#include <time.h>
#include <iostream>
#include <vector>
#include <list>
#include <string>
#include "allegro.h"
#if (USE_PNG == 1)
#include "loadpng.h"
#endif
#if (USE_JPG == 1)
#include "jpgalleg.h"
#endif
#include "enginei.h"
#include "glch_dat.h"
#include "logfile.h"


using namespace GameEngine;
using namespace std;
using LogFile::gl_log;

// Make interrupts available to C
extern "C"
{
    static void Engine_cycle_incrementor(void);
    static void Engine_one_second_timer(void);
    static void Engine_keyboard_callback(int scancode);
    static void Engine_mouse_callback(int flags);
};

static volatile unsigned int    Engine_cycle_counter;
static volatile unsigned int    Engine_last_fps;
static volatile unsigned int    Engine_frame_counter;
static volatile unsigned int    Engine_total_frames;
static volatile unsigned int    Engine_seconds;
static volatile unsigned int    Engine_idle_seconds;

Engine * Engine::s_instance = 0;

static void Engine_cycle_incrementor()
{
    Engine_cycle_counter++;
}
END_OF_FUNCTION(Engine_cycle_incrementor);

static void Engine_one_second_timer()
{
    // Keep track of last FPS
    Engine_last_fps = Engine_frame_counter;
    Engine_total_frames += Engine_frame_counter;
    Engine_frame_counter = 0;

    // Keep track of total seconds
    Engine_seconds++;

    // Keep track of idle seconds.  This will periodically be reset to
    // zero elsewhere.
    Engine_idle_seconds++;
}
END_OF_FUNCTION(Engine_one_second_timer);

static void Engine_keyboard_callback(int scancode)
{
    Engine_idle_seconds = 0;
}
END_OF_FUNCTION(Engine_keyboard_callback);

static void Engine_mouse_callback(int flags)
{
    // If there is any mouse activity (which there should be otherwise
    // this wouldn't be called... ;), reset idle count.
    if (flags != 0)
    {
        Engine_idle_seconds = 0;
    }
}
END_OF_FUNCTION(Engine_mouse_callback);

Engine *
Engine::get_instance()
{
    if (s_instance == 0)
    {
        s_instance = new Engine();
    }

    return s_instance;
}

Engine::Engine()
    throw (InitError)
{
    srand(time(NULL));
    
    log_os_type();

    gl_log->info("Installing joystick");
    /* install a joystick driver */
    if (install_joystick(JOY_TYPE_AUTODETECT) != 0) 
        gl_log->warning("Error initializing joystick");
	
    install_keyboard();
    install_timer();

    if (install_mouse() > 0)
    {
        gl_log->info("Installed mouse");
        p_is_mouse_installed = true;
    }
    else
    {
        gl_log->warning("Error initializing mouse");
        p_is_mouse_installed = false;
    }
    p_infinite_mouse_x  = 0;
    p_infinite_mouse_y  = 0;
        
    // Make Allegro aware of PNG and JPEG file formats.
#if (USE_PNG == 1)
    register_png_file_type();
#endif
#if (USE_JPG == 1)
    register_bitmap_file_type("jpg", load_jpg, NULL);
#endif

    Engine_cycle_counter = 0;
    LOCK_VARIABLE(Engine_cycle_counter);
    LOCK_FUNCTION(Engine_cycle_incrementor);
    install_int_ex(Engine_cycle_incrementor, BPS_TO_TIMER(TARGET_FPS));

    LOCK_VARIABLE(Engine_last_fps);
    LOCK_VARIABLE(Engine_frame_counter);
    LOCK_VARIABLE(Engine_total_frames);
    LOCK_VARIABLE(Engine_seconds);
    LOCK_VARIABLE(Engine_idle_seconds);
    LOCK_FUNCTION(Engine_one_second_timer);
    install_int_ex(Engine_one_second_timer, BPS_TO_TIMER(1));


    LOCK_FUNCTION(Engine_keyboard_callback);
    keyboard_lowlevel_callback = Engine_keyboard_callback;

    if (p_is_mouse_installed)
    {
        LOCK_FUNCTION(Engine_mouse_callback);
        mouse_callback = Engine_mouse_callback;
    }

    p_system_container  = new GEContainer();
    p_root_container    = new GEContainer();

    p_timer_manager = new Timer_Manager();
    p_timer_manager->adjust_frequency(updates_per_second());
    p_system_container->add_updatable(p_timer_manager,
                                      TIMER_MANAGER_UPDATE_PRIORITY);

    p_keypress_manager = 0;
    p_keypress_manager = new Keypress_Manager(this);
    p_system_container->add_updatable(p_keypress_manager,
                                      KEYPRESS_MANAGER_UPDATE_PRIORITY);
    
    p_joystick_manager = 0;
    p_joystick_manager = new Joystick_Manager(this);
    if (! p_joystick_manager->is_joystick_detected())
    {
        gl_log->warning("Joystick not detected");
    }
    else
    {
        p_system_container->add_updatable(p_joystick_manager,
                                          JOYSTICK_MANAGER_UPDATE_PRIORITY);
    }

    // Always setup Engine_Manager last as it may depend on the other
    // managers.
    p_engine_manager = new Engine_Manager(this);
    p_system_container->add_updatable(p_engine_manager,
                                      ENGINE_MANAGER_UPDATE_PRIORITY);
    p_system_container->add_drawable(p_engine_manager,
                                     ENGINE_MANAGER_DRAW_PRIORITY);

    p_null_background_task      = new Null_Background_Task();
    p_background_task           = p_null_background_task;
}

Engine::~Engine()
{
    int i;

    for (i = 0; i < NUM_FRAMES; i++)
    {
        if (p_frames[i] != 0)
            destroy_bitmap(p_frames[i]);
    }

    delete p_engine_manager;
    delete p_timer_manager;
    delete p_keypress_manager;
    delete p_joystick_manager;

    delete p_root_container;
    delete p_system_container;
    
    return;
}

void Engine::initialize_sound(bool use_software_mixer)
    throw (InitError)
{
    set_volume_per_voice(1);
#ifdef ALLEGRO_WINDOWS
    if (use_software_mixer)
    {
        if (install_sound(DIGI_DIRECTAMX(0), MIDI_NONE, NULL) != 0)
        {
            gl_log->warning("DIGI_DIRECTAMX(0) failed: " +
                            string(allegro_error));
            if (install_sound(DIGI_DIRECTAMX(1), MIDI_NONE, NULL) != 0)
            {
                throw InitError("Error initializing sound: " +
                                string(allegro_error));
            }
        }
    }
    else
#endif
    {
        if (install_sound(DIGI_AUTODETECT, MIDI_NONE, NULL) != 0)
        {
            throw InitError("Error initializing sound: " +
                            string(allegro_error));
        }
    }
    
    log_digi_card();
    // Max out volume
    set_volume(255, -1);
}


void
Engine::set_graphics_mode(int width, int height, int color_depth,
                          bool use_fullscreen)
    throw (InitError)
{
    int mode;

    if (use_fullscreen)
    {
        mode = GFX_AUTODETECT_FULLSCREEN;
    }
    else
    {
        mode = GFX_AUTODETECT_WINDOWED;
    }
    
    set_color_depth(color_depth);
    if (set_gfx_mode(mode, width, height, 0, 0) < 0)
    {
        // If the bit depth was 16, try 15 before giving up.
        if (color_depth == 16)
        {
            gl_log->warning("Trying 15 BPP color depth, 16 BPP failed: " +
                            string(allegro_error));
            set_color_depth(15);
            if (set_gfx_mode(mode, width, height, 0, 0) < 0)
            {
                // Ok, really give up.
                throw InitError("Could not change graphics mode: " +
                                string(allegro_error));
            }
        }
        else
        {
            throw InitError("Could not change graphics mode: " +
                            string(allegro_error));
        }
    }

    gl_log->info(string("Graphics driver: ") + gfx_driver->name);
    gl_log->info(string("Description: ") + gfx_driver->desc);
}

void
Engine::set_animation_mode(Animation_Method animation_method, bool use_vram,
                           bool use_vsync)
    throw (InitError)
{
    p_frames[0] = p_frames[1] = 0;
    p_use_vram  = use_vram;
    p_use_vsync = use_vsync;

    p_frames[0] = create_video_bitmap(SCREEN_W, SCREEN_H);
    if (p_frames[0] == 0)
    {
        throw InitError("Could not grab screen");
    }
    clear(p_frames[0]);

    // Try page flipping, if requested
    if (animation_method == PAGE_FLIPPING)
        p_frames[1] = create_video_bitmap(SCREEN_W, SCREEN_H);

    if (p_frames[1] == 0)
    {
        // Default to double buffering
        p_frames[1] = create_bitmap(SCREEN_W, SCREEN_H);
        if (p_frames[1] == 0)
            throw InitError("Could not double buffer");
        clear(p_frames[1]);
        p_current_frame = 1;
        show_video_bitmap(p_frames[0]);
        p_animation_method = DOUBLE_BUFFERING;
        gl_log->info("Using double buffering");
        gl_log->info(string("Vsync is ") +
                  ((p_use_vsync) ? "enabled" : "disabled"));
    }
    else
    {
        // Got 2 pages, use page flipping
        clear(p_frames[1]);
        show_video_bitmap(p_frames[1]);
        p_animation_method = PAGE_FLIPPING;
        p_current_frame = 0;
        gl_log->info("Using page flipping");
    }

    p_frame = p_frames[p_current_frame];
}

void
Engine::poll_background_task()
{
    p_background_task->poll();
}

void
Engine::catchup()
{
    p_actual_cycle      = Engine_cycle_counter;
}

void
Engine::run()
{
    Engine_frame_counter        = 0;
    Engine_cycle_counter        = 0;
    Engine_total_frames         = 0;
    Engine_last_fps             = 0;
    Engine_seconds              = 0;
    Engine_idle_seconds         = 0;

    p_end_game  = false;
    p_paused    = false;
    p_actual_cycle      = 0;
    p_loop_count        = 0;

    gl_log->info("Starting engine loop");
    
    p_background_task->resume();
    do
    {
        p_catchup = 0;
        p_background_task->poll();

        while(Engine_cycle_counter <= p_actual_cycle)
            p_background_task->poll();

        while(Engine_cycle_counter > p_actual_cycle)
        {
            do_one_cycle();
            if (p_catchup > 0)
                p_background_task->poll();
            p_catchup++;
        }

        get_keyboard_input();
        
        draw_one_frame();

        p_loop_count++;
    } while (! p_end_game);
    p_background_task->suspend();

    // Damn ostrstream bug again!
    char            buf[81];
    memset(buf, 0, 81);
    ostrstream   fps_string(buf, 80);

    fps_string << "Average FPS: " << get_average_frames_per_second();
    gl_log->info(fps_string.str());

    return;
}

void
Engine::stop()
{
    p_end_game = true;
}

GEContainer *
Engine::get_root_container()
{
    return p_root_container;
}

void
Engine::set_background_task(Background_Task * bg_task)
{
    if (bg_task != 0)
        p_background_task = bg_task;
    else
        p_background_task = p_null_background_task;
    return;
}

unsigned int
Engine::updates_per_second()
{
    return TARGET_FPS;
}

unsigned int
Engine::get_frames_per_second()
{
    return Engine_last_fps;
}

int
Engine::get_catchup()
{
    return p_catchup;
}

unsigned int
Engine::get_idle_timer()
{
    return Engine_idle_seconds;
}

void
Engine::reset_idle_timer()
{
    Engine_idle_seconds = 0;
}

void
Engine::get_keyboard_input()
{
    if ((key[KEY_D]) && (key[KEY_A]) && (key[KEY_V]))
    {
        gl_log->info("Backdoor exit!");
	p_end_game = true;
    }

    if (key[KEY_F12])
    {
        BITMAP *    shot;
        PALETTE     pal;

        shot = create_bitmap(SCREEN_W, SCREEN_H);
        blit(screen, shot, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
        save_bitmap("shot.png", shot, pal);
    }

    if (key[KEY_P])
        p_paused = true;
    if (! key[KEY_P])
        p_paused = false;
}

void
Engine::draw_one_frame()
{
    // Always call system container after the root container to allow
    // system ojbects the highest draw priority.
    p_root_container->draw(p_frame);
    p_system_container->draw(p_frame);

    if (p_animation_method == PAGE_FLIPPING)
    {
        show_video_bitmap(p_frame);
        p_current_frame++;
        p_current_frame %= NUM_FRAMES;
        p_frame = p_frames[p_current_frame];
    }
    else
    {
        if (p_use_vsync)
            vsync();
        blit(p_frame, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
    }
    Engine_frame_counter++;
}

void
Engine::do_one_cycle()
{
    if (! p_paused)
    {
        // Always update system objects before user level objects.
        (void) p_system_container->update();
        (void) p_root_container->update();
    }

    if (p_is_mouse_installed)
    {
        int delta_x;
        int delta_y;

        get_mouse_mickeys(&delta_x, &delta_y);
        p_infinite_mouse_x  += delta_x;
        p_infinite_mouse_y  += delta_y;
    }
        
    p_actual_cycle++;

    return;
}

float
Engine::get_average_frames_per_second()
{
    return float(Engine_total_frames) / float(Engine_seconds);
}


unsigned int
Engine::elapsed_seconds()
{
    return Engine_seconds;
}

Timer_Factory *
Engine::get_timer_factory()
{
    return p_timer_manager;
}

Keypress_Factory *
Engine::get_keypress_factory()
{
    return p_keypress_manager;
}

Joystick_Factory *
Engine::get_joystick_factory()
{
    return p_joystick_manager;
}


void
Engine::log_os_type()
{
    string prefix;

    prefix = "OS type: ";

    // os_type gets set in allegro_init().

    switch (os_type)
    {
        case OSTYPE_WIN3:
            gl_log->info(prefix + "Windows 3.1 or earlier");
            break;

        case OSTYPE_WIN95:
            gl_log->info(prefix + "Windows 95");
            break;

        case OSTYPE_WIN98:
            gl_log->info(prefix + "Windows 98");
            break;

        case OSTYPE_WINME:
            gl_log->info(prefix + "Windows ME");
            break;

        case OSTYPE_WINNT:
            gl_log->info(prefix + "Windows NT");
            break;

        case OSTYPE_WIN2000:
            gl_log->info(prefix + "Windows 2000");
            break;

        case OSTYPE_WINXP:
            gl_log->info(prefix + "Windows XP");
            break;

        case OSTYPE_OS2:
            gl_log->info(prefix + "OS/2");
            break;

        case OSTYPE_WARP:
            gl_log->info(prefix + "OS/2 Warp 3");
            break;

        case OSTYPE_DOSEMU:
            gl_log->info(prefix + "Linux DOSEMU");
            break;

        case OSTYPE_OPENDOS:
            gl_log->info(prefix + "Caldera OpenDOS");
            break;

        case OSTYPE_LINUX:
            gl_log->info(prefix + "Linux");
            break;

        case OSTYPE_FREEBSD:
            gl_log->info(prefix + "FreeBSD");
            break;

        case OSTYPE_UNIX:
            gl_log->info(prefix + "Unknown Unix variant");
            break;

        case OSTYPE_UNKNOWN:
            gl_log->info(prefix + "Unknown, or regular MSDOS");
            break;

        default:
            gl_log->info(prefix + "Mystery OS");
            break;
    }
    gl_log->info(string("Platform: ") + ALLEGRO_PLATFORM_STR);

    return;
}

void
Engine::log_digi_card()
{
    string prefix;

    prefix = "Sound card type: ";

    // digi_card gets set in install_sound().

    switch (digi_card)
    {
#ifdef ALLEGRO_DOS
        case DIGI_SB10:
            gl_log->info(prefix + "SB 1.0 (8 bit mono single shot dma)");
            break;

            
        case DIGI_SB15:
            gl_log->info(prefix + "SB 1.5 (8 bit mono single shot dma)");
            break;

        case DIGI_SB20:
            gl_log->info(prefix + "SB 2.0 (8 bit mono auto-initialised dma)");
            break;

        case DIGI_SBPRO:
            gl_log->info(prefix + "SB Pro (8 bit stereo)");
            break;

        case DIGI_SB16:
            gl_log->info(prefix + "SB16 (16 bit stereo)");
            break;
            
        case DIGI_AUDIODRIVE:
            gl_log->info(prefix + "ESS AudioDrive");
            break;

        case DIGI_SOUNDSCAPE:
            gl_log->info(prefix + "Ensoniq Soundscape");
            break;
            
        case DIGI_WINSOUNDSYS:
            gl_log->info(prefix + "Windows Sound System");
            break;
#endif // ALLEGRO_DOS

#ifdef ALLEGRO_UNIX
        case DIGI_OSS:
            gl_log->info(prefix + "Open Sound System");
            break;

        case DIGI_ESD:
            gl_log->info(prefix + "Enlightend Sound Daemon");
            break;

#ifdef DIGI_ALSA
        case DIGI_ALSA:
            gl_log->info(prefix + "ALSO sound driver");
            break;
#endif

#endif // ALLEGRO_UNIX

#ifdef ALLEGRO_WINDOWS
        case DIGI_DIRECTX(0):
            gl_log->info(prefix + "DirectSound, device 0");
            break;
        case DIGI_DIRECTX(1):
            gl_log->info(prefix + "DirectSound, device 1");
            break;
        case DIGI_DIRECTAMX(0):
            gl_log->info(prefix + "DirectSound with Allegro mixing, device 0");
            break;
        case DIGI_DIRECTAMX(1):
            gl_log->info(prefix + "DirectSound with Allegro mixing, device 1");
            break;
#endif // ALLEGRO_WINDOWS

        default:
            gl_log->info(prefix + "Mystery sound card");
            break;
    }

    return;
}
        
int
Engine::get_infinite_mouse_x()
{
    return p_infinite_mouse_x;
}

void
Engine::set_infinite_mouse_x(int infinite_mouse_x)
{
    p_infinite_mouse_x = infinite_mouse_x;
}

int
Engine::get_infinite_mouse_y()
{
    return p_infinite_mouse_y;
}

void
Engine::set_infinite_mouse_y(int infinite_mouse_y)
{
    p_infinite_mouse_y = infinite_mouse_y;
}

InitError::InitError(const string & error_string)
{
    p_error_string = error_string;
}

const string &
InitError::error_string()
{
    return p_error_string;
}


GEContainer::GEContainer()
    : p_drawable_objects(), p_updatable_objects()
{
}

GEContainer::~GEContainer()
{
}

void
GEContainer::draw(BITMAP * bitmap)
{
    Drawable_List::iterator     i;

    for (i = p_drawable_objects.begin(); i != p_drawable_objects.end(); i++)
    {
        i->second->draw(bitmap);
    }
}

bool
GEContainer::update()
{
    Updatable_List::iterator    i;
    bool        updated_any;
    bool        updated_last;
    
    updated_any = false;
    for (i = p_updatable_objects.begin(); i != p_updatable_objects.end(); i++)
    {
        updated_last    = i->second->update();
        updated_any     |= updated_last;
    }

    return updated_any;
}


void
GEContainer::add_drawable(Drawable * object, int priority)
{
    p_drawable_objects.insert(make_pair(priority, object));
}

void
GEContainer::remove_drawable(Drawable * object)
{
    Drawable_List::iterator first;
    Drawable_List::iterator last;
    Drawable_List::iterator i;

    first       = p_drawable_objects.begin();
    last        = p_drawable_objects.end();

    for (i = first; i != last; i++)
    {
        if (i->second == object)
        {
            p_drawable_objects.erase(i);
        }
    }
}

void
GEContainer::add_updatable(Updatable * object, int priority)
{
    p_updatable_objects.insert(make_pair(priority, object));
}

void
GEContainer::remove_updatable(Updatable * object)
{
    Updatable_List::iterator first;
    Updatable_List::iterator last;
    Updatable_List::iterator i;

    first       = p_updatable_objects.begin();
    last        = p_updatable_objects.end();

    for (i = first; i != last; i++)
    {
        if (i->second == object)
        {
            p_updatable_objects.erase(i);
        }
    }
}


void
Null_Background_Task::poll()
{
}

void
Null_Background_Task::resume()
{
}

void
Null_Background_Task::suspend()
{
}

/************************************************************************
 *
 * Engine_Manager
 *
 ************************************************************************/

Engine_Manager::Engine_Manager(Engine * engine)
{
    p_engine    = engine;
    p_fps_toggle_key    =
        p_engine->get_keypress_factory()->create_keypress_watcher(KEY_F11);
    p_fps_toggle_key->set_one_time();

    p_debug_toggle_key  =
        p_engine->get_keypress_factory()->create_keypress_watcher(KEY_F10);
    p_debug_toggle_key->set_one_time();
    
    p_show_fps      = false;
    p_show_debug    = false;
}

Engine_Manager::~Engine_Manager()
{
    delete p_debug_toggle_key;
    delete p_fps_toggle_key;
}

void
Engine_Manager::toggle_flag(bool & flag)
{
    if (flag == true)
        flag = false;
    else // flag == false
        flag = true;
}

bool
Engine_Manager::update()
{
    bool    updated;

    updated = false;
    if (p_fps_toggle_key->is_pressed())
    {
        toggle_flag(p_show_fps);
        updated = true;
    }

    if (p_debug_toggle_key->is_pressed())
    {
        toggle_flag(p_show_debug);
    }

    return updated;
}

void
Engine_Manager::draw(BITMAP * frame)
{
    int current_line;
    int black;
    int white;

    black   = makecol(0, 0, 0);
    white   = makecol(255, 255, 255);
    text_mode(black);
        
    current_line = 0;

    if (p_show_fps)
    {
        textprintf(frame, font, 0, current_line, white, "FPS: %2d",
                   p_engine->get_frames_per_second());
        current_line += text_height(font);
    }

    if (p_show_debug)
    {
        textprintf(frame, font, 0, current_line, white, "Idle time: %d",
                   p_engine->get_idle_timer());
        current_line += text_height(font);
    }
}
