#ifndef _ENGINEI_H_
#define _ENGINEI_H_

#include "engine.h"

namespace GameEngine
{
    /**
     * This allows the engine to provide some house keeping functions
     * such as display the current FPS, save screen shots, etc.
     */
    class Engine_Manager : public Updatable, public Drawable
    {
      public:
        /**
         * Construct a new engine manager.
         *
         * @param engine The game engine.
         */
        Engine_Manager(Engine * engine);

        /**
         * Destructor;
         */
        virtual ~Engine_Manager();

        virtual bool update();

        virtual void draw(BITMAP * frame);

      private:
        /**
         * Toggles the value of a boolean flag.
         *
         * @param flag Flag to toggle.
         */
        void toggle_flag(bool & flag);

        /** Current game engine. */
        Engine *            p_engine;

        /** Watches for FPS toggle key. */
        Keypress_Watcher *  p_fps_toggle_key;

        /** Watches for debug toggle key. */
        Keypress_Watcher *  p_debug_toggle_key;

        /** True if currently showing FPS. */
        bool                p_show_fps;

        /** True if currently showing debug information. */
        bool                p_show_debug;
    };
    
    class Timer_Impl : public Timer
    {
      public:
        virtual void    set(unsigned int milliseconds);
        virtual void    set_interval(unsigned int milliseconds);
        virtual void    disable();
        virtual bool    expired();
        virtual bool    stopped();
        virtual         ~Timer_Impl();

        // Operations used by Timer_Manager
        Timer_Impl(Timer_Manager * manager, int ticks_per_second);
        void    tick_tock();
        void    adjust_frequency(unsigned int ticks_per_second);

      private:
        enum State { STOPPED, RUNNING };

        Timer_Manager * p_timer_manager;
        unsigned int    p_interval;
        unsigned int    p_ticks_left;
        unsigned int    p_ticks_per_second;
        State           p_state;
    };

    class Timer_Manager : public Updatable, public Timer_Factory
    {
      public:
        Timer_Manager();
        
        virtual Timer * create_timer();
        void            delete_timer(Timer * timer);
        virtual bool    update();
        void            adjust_frequency(unsigned int ticks_per_second);

      private:
        typedef list<Timer_Impl *>   Timer_List_Impl;

        unsigned int    p_ticks_per_second;
        Timer_List_Impl p_timers;
    };

    class Keypress_Watcher_Impl : public Keypress_Watcher
    {
      public:
        virtual void    set_scancode(int scancode);
        virtual bool    is_pressed();
        virtual bool    is_artificial();
        virtual bool    is_physically_pressed();
        virtual void    set_repeat(int initial_delay, int repeat_delay);
        virtual void    set_continuous();
        virtual void    set_one_time();
        void            watch();

        Keypress_Watcher_Impl(Keypress_Manager * keypress_manager,
                              Timer * timer);
        virtual ~Keypress_Watcher_Impl();

      private:
        void    reset_state();

        enum Type {REPEAT, CONTINUOUS, ONE_TIME};

        Type    p_type;
        int     p_scancode;
        bool    p_is_artificial;
        bool    p_is_pressed;
        bool    p_was_pressed;
        int     p_initial_delay;
        int     p_repeat_delay;
        Timer * p_timer;
        Keypress_Manager *      p_keypress_manager;
    };

    class Keypress_Manager : public Keypress_Factory, public Updatable
    {
      public:
        Keypress_Watcher *      create_keypress_watcher(int scancode);
        void                    delete_keypress_watcher(Keypress_Watcher *
                                                        keypress_watcher);
        bool                    update();

        Keypress_Manager(Engine * engine);
        virtual ~Keypress_Manager();

      private:
        typedef list<Keypress_Watcher_Impl *>   Keypress_Watcher_List;

        Keypress_Watcher_List   p_keypress_watcher_list;
        Engine *        p_engine;
    };
    
    class Joystick_Watcher_Impl : public Joystick_Watcher
    {
      public:
        virtual void    set_joystick_event(Joystick_Event * event);
        virtual bool    is_pressed();
        virtual void    set_repeat(int initial_delay, int repeat_delay);
        virtual void    set_continuous();
        virtual void    set_one_time();
        void            watch();

        Joystick_Watcher_Impl(Joystick_Manager * joystick_manager,
                              Timer * timer, Joystick_Event * event);
        virtual ~Joystick_Watcher_Impl();

      private:
        void    reset_state();

        enum Type {REPEAT, CONTINUOUS, ONE_TIME};
 	 	
        Type    p_type;
        Joystick_Event *    p_event;
        bool    p_is_pressed;
        bool    p_was_pressed;
        int     p_initial_delay;
        int     p_repeat_delay;
        Timer * p_timer;
        Joystick_Manager *      p_joystick_manager;
    };

    class Null_Joystick_Watcher_Impl : public Joystick_Watcher
    {
      public:
        virtual void    set_joystick_event(Joystick_Event * event);
        virtual bool    is_pressed();
        virtual void    set_repeat(int initial_delay, int repeat_delay);
        virtual void    set_continuous();
        virtual void    set_one_time();
    };
    
    class Joystick_Manager : public Joystick_Factory, public Updatable
    {
      public:
        Joystick_Watcher *      create_joystick_watcher(Joystick_Event *
                                                        event);
        bool                    is_joystick_detected();
        
        void                    delete_joystick_watcher(Joystick_Watcher *
                                                        joystick_watcher);
        bool                    update();

        Joystick_Manager(Engine * engine);
        virtual ~Joystick_Manager();

      private:
        typedef list<Joystick_Watcher_Impl *>   Joystick_Watcher_List;

        Joystick_Watcher_List   p_joystick_watcher_list;
        Engine *        p_engine;
    };
}
#endif

/* Local Variables: */
/* mode: c++ */
/* End: */
