#include "enginei.h"
#include "logfile.h"


using namespace GameEngine;
using namespace std;
using LogFile::gl_log;


Joystick_Watcher_Impl::Joystick_Watcher_Impl(
    Joystick_Manager * joystick_manager,
    Timer * timer,
    Joystick_Event * event)
{
    p_type      = CONTINUOUS;
    p_event             = event;
    p_is_pressed        = false;
    p_was_pressed       = false;
    p_initial_delay     = 0;
    p_repeat_delay      = 0;
    p_timer             = timer;
    p_joystick_manager  = joystick_manager;
}     

Joystick_Watcher_Impl::~Joystick_Watcher_Impl()
{
    delete p_timer;
    p_timer = 0;

    delete p_event;
    p_event = 0;

    p_joystick_manager->delete_joystick_watcher(this);
}

void
Joystick_Watcher_Impl::set_joystick_event(Joystick_Event * event)
{
    delete p_event;
    p_event     = event;
}

bool
Joystick_Watcher_Impl::is_pressed()
{
    return p_is_pressed;
}

void
Joystick_Watcher_Impl::watch()
{
    bool    joystick_activity = false;
   
    joystick_activity = p_event->is_triggered();

    if ((! joystick_activity) && (!p_was_pressed))
       	return;
        	
    if (! joystick_activity)
    {
        reset_state();
        return;
    }
      	
    if (p_type == CONTINUOUS)
        p_is_pressed        = true;
    else if (p_type == ONE_TIME)
    {
        if (! p_was_pressed)
             p_is_pressed    = true;
        else        
             p_is_pressed    = false;
    }
    else if (p_type == REPEAT)
    {
        if (! p_was_pressed)
        {
            // Initial joystick
            p_is_pressed    = true;
            p_timer->set(p_initial_delay);
        }
        else if (p_timer->expired())
        {
            p_is_pressed    = true;
            p_timer->set(p_repeat_delay);
        }
        else
            p_is_pressed    = false;
    }
    p_was_pressed = true;
    return;
}

void
Joystick_Watcher_Impl::set_continuous()
{
    reset_state();
    p_type      = CONTINUOUS;
}

void
Joystick_Watcher_Impl::set_one_time()
{
    reset_state();
    p_type      = ONE_TIME;
 
}

void
Joystick_Watcher_Impl::set_repeat(int initial_delay, int repeat_delay)
{
    reset_state();
    p_initial_delay     = initial_delay;
    p_repeat_delay      = repeat_delay;
    p_type              = REPEAT;
}

void
Joystick_Watcher_Impl::reset_state()
{
    p_is_pressed    = false;
    p_was_pressed   = false;
    if (p_type == REPEAT)
        p_timer->disable();
}


void
Null_Joystick_Watcher_Impl::set_joystick_event(Joystick_Event * event)
{
}

bool
Null_Joystick_Watcher_Impl::is_pressed()
{
    return false;
}

void
Null_Joystick_Watcher_Impl::set_repeat(int initial_delay, int repeat_delay)
{
}

void
Null_Joystick_Watcher_Impl::set_continuous()
{
}

void
Null_Joystick_Watcher_Impl::set_one_time()
{
}



Joystick_Manager::Joystick_Manager(Engine * engine)
{
    p_engine = engine;
}

Joystick_Manager::~Joystick_Manager()
{
}

Joystick_Watcher *
Joystick_Manager::create_joystick_watcher(Joystick_Event * event)
{
    if (event == 0)
    {
        return new Null_Joystick_Watcher_Impl();
    }
    else
    {
        Joystick_Watcher_Impl * joystick_watcher;
        Timer *                 timer;

        timer   = p_engine->get_timer_factory()->create_timer();
        joystick_watcher    = new Joystick_Watcher_Impl(this, timer, event);
        p_joystick_watcher_list.push_back(joystick_watcher);

        return joystick_watcher;
    }
}

bool
Joystick_Manager::is_joystick_detected()
{
    return (num_joysticks > 0);
}

void
Joystick_Manager::delete_joystick_watcher(Joystick_Watcher * joystick_watcher)
{
    Joystick_Watcher_List::iterator     i;

    for (i =  p_joystick_watcher_list.begin();
         i != p_joystick_watcher_list.end();
         i++)
    {
        if (*i == joystick_watcher)
        {
            p_joystick_watcher_list.erase(i);
            break;
        }
    }
    return;
}

bool
Joystick_Manager::update()
{
    Joystick_Watcher_List::iterator     i;

    poll_joystick();
    for (i =  p_joystick_watcher_list.begin();
         i != p_joystick_watcher_list.end();
         i++)
    {
 
       (*i)->watch();
    }
    // No visual updates
    return false;
}

Joystick_Button_Event::Joystick_Button_Event(int joystick, int button)
{
    // Eh, we should probably throw an exception here, but for now,
    // just silently corret the problem.
    if (joystick >= (num_joysticks -1))
        joystick = num_joysticks - 1;

    if (button >= (joy[joystick].num_buttons - 1))
        button = joy[joystick].num_buttons - 1;

    p_joystick  = joystick;
    p_button    = button;
}

bool
Joystick_Button_Event::is_triggered()
{
    return (joy[p_joystick].button[p_button].b);
}

Joystick_Digital_Axis_Event::Joystick_Digital_Axis_Event(
    int joystick, int stick, int axis, Position position)
{
    // Eh, we should probably throw an exception here, but for now,
    // just silently corret the problem.
    if (joystick >= (num_joysticks -1))
        joystick = num_joysticks - 1;

    if (stick >= (joy[joystick].num_sticks - 1))
        stick = joy[joystick].num_sticks - 1;

    if (axis >= (joy[joystick].stick[stick].num_axis -1))
        axis = joy[joystick].stick[stick].num_axis -1;

    p_joystick  = joystick;
    p_stick     = stick;
    p_axis      = axis;
    p_position  = position;
}

bool
Joystick_Digital_Axis_Event::is_triggered()
{
    bool    triggered;
    
    switch(p_position)
    {
        case POSITION_ONE:
            triggered = joy[p_joystick].stick[p_stick].axis[p_axis].d1;
            break;

        case POSITION_TWO:
            triggered = joy[p_joystick].stick[p_stick].axis[p_axis].d2;
            break;

        default:
            triggered = false;
            break;
    }

    return triggered;
}
