#include "gui/horiz.h"
#include "gui/uimgr.h"
#include "outline.h"
#include <list>

using namespace GUI;
using namespace Outline;


/************************************************************************
 *
 * Horizontal_UI_Factory
 *
 ************************************************************************/

Label_UI *
Horizontal_UI_Factory::create_label_ui()
{
    return new Horizontal_Label_UI();
}

List_UI *
Horizontal_UI_Factory::create_list_ui()
{
    return new Horizontal_List_UI();
}

Menu_UI *
Horizontal_UI_Factory::create_menu_ui()
{
    return new Horizontal_Menu_UI();
}

Separator_UI *
Horizontal_UI_Factory::create_separator_ui()
{
    return new Horizontal_Separator_UI();
}


/************************************************************************
 *
 * Horizontal_Label_UI
 *
 ************************************************************************/

Horizontal_Label_UI::Horizontal_Label_UI()
    : Abstract_Label_UI()
{
}

void
Horizontal_Label_UI::install_ui(Label * label)
{
    int width;
    int height;

    destroy_bitmaps();

    p_label = label;
    width = p_label->get_maximum_width();
    height = outline_text_height(p_label->get_font());
    p_main_bitmap = ::create_bitmap(width, height);
    label_updated();
}

void
Horizontal_Label_UI::label_updated()
{
    p_label_bitmap = draw_horizontal(SUB_BITMAP, p_main_bitmap->w,
                                     p_main_bitmap->h);
    p_label->set_size(p_label_bitmap->w, p_label_bitmap->h);
}


/************************************************************************
 *
 * Horizontal_List_UI
 *
 ************************************************************************/

void
Horizontal_List_UI::install_list(List * list)
{
    unsigned i;

    p_list = list;
    p_x = p_list->get_x();
    p_y = p_list->get_y();
    p_width = p_list->get_width();
    p_height = p_list->get_height();

    p_num_visible_items = p_height / p_item_height;

    // Adjust height to be a multiple of p_item_height.  This is
    // necessary so that it clips exactly on an item boundary.
    p_height = p_num_visible_items * p_item_height;

    // Create the visible text items
    for (i = 0; i < p_num_visible_items; i++)
    {
        Label * label;

        label = new Label(p_width, p_item_height);
        p_visible_items.push_back(label);
    }

    // Create the off screen text items
    for (i = 0; i < p_num_visible_items; i++)
    {
        Label * label;

        label = new Label(p_width, p_item_height);
        p_offscreen_items.push_back(label);

    }
    // Add items to the container
    Label_List::iterator      j;

    for (j = p_visible_items.begin(); j != p_visible_items.end(); j++)
    {
        p_container->add_drawable((*j), DRAW_PRIORITY);
        p_container->add_updatable((*j), UPDATE_PRIORITY);
    }

    for (j = p_offscreen_items.begin(); j != p_offscreen_items.end(); j++)
    {
        p_container->add_drawable((*j), DRAW_PRIORITY);
        p_container->add_updatable((*j), UPDATE_PRIORITY);
    }

    hide_split(0);
}

int
Horizontal_List_UI::get_justification_offset(Label * label)
{
    int offset;
    
    switch (p_justification)
    {
        case CENTER_JUSTIFIED:
            offset = (p_width / 2) - (label->get_width() / 2);
            break;

        case RIGHT_JUSTIFIED:
            offset = p_width - label->get_width();
            break;

        case LEFT_JUSTIFIED:
        default:
            offset = 0;
    }

    return offset;
}

void
Horizontal_List_UI::hide_item_left(Label * item, int index, int milliseconds)
{
    int x;
    int y;

    x = p_x - p_width + get_justification_offset(item);
    y = p_list->get_y() + (p_item_height * index);

    item->set_location(x, y, milliseconds);
}

void
Horizontal_List_UI::hide_item_right(Label * item, int index, int milliseconds)
{
    int x;
    int y;

    x = p_x + p_width + get_justification_offset(item);
    y = p_list->get_y() + (p_item_height * index);

    item->set_location(x, y, milliseconds);
}

void
Horizontal_List_UI::hide_item_top(Label * item, int index, int milliseconds)
{
    int x;
    int y;

    x = p_x + get_justification_offset(item);
    y = p_list->get_y() + (p_item_height * index) - p_height;

    item->set_location(x, y, milliseconds);
}

void
Horizontal_List_UI::center_item(Label * item, int milliseconds)
{
    int x;
    int y;

    x = p_x + get_justification_offset(item);
    y = (p_list->get_y() + p_list->get_height() / 2) - (p_item_height / 2);

    item->set_location(x, y, milliseconds);
}

void
Horizontal_List_UI::unhide_item(Label * item, int index, int milliseconds)
{
    int x;
    int y;

    x = p_x + get_justification_offset(item);
    y = p_list->get_y() + (p_item_height * index);

    item->set_location(x, y, milliseconds);
}

void
Horizontal_List_UI::hide_offscreen_item(Label * item)
{
    int x;
    int y;

    // Offscreen items are "parked" just above the list
    x = p_x;
    y = p_y - p_item_height;

    item->set_location(x, y, 0);
}

/************************************************************************
 *
 * Horizontal_Menu_UI
 *
 ************************************************************************/

void
Horizontal_Menu_UI::install_menu(Menu * menu)
{
    int x;
    int y;
    int width;
    int height;
    int title_height;
    int y_offset;
        
    p_menu = menu;
    x = p_menu->get_x();
    y = p_menu->get_y();
    width = p_menu->get_width();
    height = p_menu->get_height();
    y_offset = 0;
    
    title_height = 2 * p_item_height;
    p_title_list = new List(x, y, width, title_height);
    p_title_model = new Menu_Title_List_Adapter(p_menu);
    p_title_list->set_model(p_title_model);
    p_title_list->hide_top(0);
    y_offset += title_height;

    p_separator = new Separator(x, y+y_offset, width, SEPARATOR_HEIGHT);
    y_offset += SEPARATOR_HEIGHT;

    p_menu_list = new List(x, y + y_offset, width, height - y_offset);
}


/************************************************************************
 *
 * Horizontal_Separator_UI
 *
 ************************************************************************/

int
Horizontal_Separator_UI::get_inner_x()
{
    int x_offset;

    switch (p_separator->get_justification())
    {
        case LEFT_JUSTIFIED:
            x_offset = 0;
            break;

        case RIGHT_JUSTIFIED:
            x_offset = p_bounding_box->get_width() / 2;
            break;

        case CENTER_JUSTIFIED:
        default:
            x_offset = p_bounding_box->get_width() / 4;
            break;
    }

    return (p_bounding_box->get_x() + x_offset);
}

int
Horizontal_Separator_UI::get_inner_y()
{
    return (p_bounding_box->get_y() + p_bounding_box->get_height() / 4);
}
