#include "outline.h"
#include "gui/list.h"
#include "gui/label.h"
#include "gui/uimgr.h"
#include "logfile.h"

using namespace Outline;
using namespace GUI;
using LogFile::gl_log;


/************************************************************************
 *
 * List_Model
 *
 ************************************************************************/

List_Model::~List_Model()
{
}


/************************************************************************
 *
 * Empty_List_Model
 *
 ************************************************************************/

class Empty_List_Model : public List_Model
{
  public:
    virtual string get_item_text_at(unsigned index) const { return ""; }
    virtual Icon * get_item_icon_at(unsigned index) const { return 0; }
    virtual unsigned get_size() const { return 0; }
};

/************************************************************************
 *
 * List
 *
 ************************************************************************/

List::List(int x, int y, int width, int height)
{
    p_x = x;
    p_y = y;
    p_width = width;
    p_height = height;
    p_justification = CENTER_JUSTIFIED;

    p_model = new Empty_List_Model();;
    p_list_changed = true;

    p_selected_text_color       = makecol(  0,   0,   0);
    p_selected_outline_color    = makecol(255, 255, 255);
    p_unselected_text_color     = makecol(255, 255, 255);
    p_unselected_outline_color  = makecol(  0,   0,   0);

    p_first_shown_index = 0;
    p_selected_index = 0;

    p_ui = UI_Manager::get_list_ui();
    p_ui->install_list(this);
    p_num_visible_items = p_ui->get_number_of_visible_items();
}

List::List(const Rectangle & bounding_box)
{
    p_x = bounding_box.get_x();
    p_y = bounding_box.get_y();
    p_width = bounding_box.get_width();
    p_height = bounding_box.get_height();
    p_justification = CENTER_JUSTIFIED;

    p_model = new Empty_List_Model();;
    p_list_changed = true;

    p_selected_text_color       = makecol(  0,   0,   0);
    p_selected_outline_color    = makecol(255, 255, 255);
    p_unselected_text_color     = makecol(255, 255, 255);
    p_unselected_outline_color  = makecol(  0,   0,   0);

    p_first_shown_index = 0;
    p_selected_index = 0;

    p_ui = UI_Manager::get_list_ui();
    p_ui->install_list(this);
    p_num_visible_items = p_ui->get_number_of_visible_items();
}

List::~List()
{
    delete p_model;
    delete p_ui;
}

int
List::get_x() const
{
    return p_x;
}

int
List::get_y() const
{
    return p_y;
}

int
List::get_width() const
{
    return p_width;
}

int
List::get_height() const
{
    return p_height;
}

void
List::set_model(const List_Model * model)
{
    delete p_model;
    p_model = model;
    p_list_changed = true;
}

const List_Model *
List::get_model() const
{
    return p_model;
}

void
List::set_justification(Justification justification)
{
    p_justification = justification;
    p_list_changed = true;
}

Justification
List::get_justification() const
{
    return p_justification;
}

void
List::set_selection_position(unsigned first_shown_index,
                             unsigned selected_index)
{
    unsigned last_index;

    // Set indices to zero if the model is empty.
    if (p_model->get_size() == 0)
    {
        p_first_shown_index = 0;
        p_selected_index = 0;
    }
    else
    {
        last_index = p_model->get_size() - 1;

        p_first_shown_index = first_shown_index;
        if (p_first_shown_index > last_index)
            p_first_shown_index = last_index;

        p_selected_index = selected_index;
        if (p_selected_index > last_index)
            p_selected_index = last_index;

        // Make sure the selected index is one of the visible items,
        // and correct if not.
        if (p_selected_index < p_first_shown_index)
            p_selected_index = p_first_shown_index;

        if (p_selected_index > (p_first_shown_index + p_num_visible_items - 1))
            p_selected_index = p_first_shown_index + p_num_visible_items - 1;
        
    }

    p_list_changed = true;
}

unsigned
List::get_selected_index() const
{
    return p_selected_index;
}

unsigned
List::get_first_shown_index() const
{
    return p_first_shown_index;
}

void
List::set_selected_colors(int text, int outline)
{
    p_selected_text_color = text;
    p_selected_outline_color = outline;
    p_list_changed = true;
}

int
List::get_selected_text_color() const
{
    return p_selected_text_color;
}

int
List::get_selected_outline_color() const
{
    return p_selected_outline_color;
}

void
List::set_unselected_colors(int text, int outline)
{
    p_unselected_text_color = text;
    p_unselected_outline_color = outline;
    p_list_changed = true;
}

int
List::get_unselected_text_color() const
{
    return p_unselected_text_color;
}

int
List::get_unselected_outline_color() const
{
    return p_unselected_outline_color;
}

void
List::line_up(Scroll_Options scroll_options)
{
    unsigned move_up;
    unsigned scroll_up;

    if (! p_ui->is_unhidden())
        return;
    
    move_up = 1;
    if (p_selected_index == p_first_shown_index)
        scroll_up = 1;
    else
        scroll_up = 0;

    move_selection_up(move_up, scroll_up, scroll_options);
}

void
List::page_up(Scroll_Options scroll_options)
{
    unsigned move_up;
    unsigned scroll_up;

    if (! p_ui->is_unhidden())
        return;
    
    if (p_selected_index == p_first_shown_index)
    {
        // At top, scroll up a page.
        move_up = p_num_visible_items - 1;
        scroll_up = p_num_visible_items - 1;
    }
    else
    {
        // Move to first shown.
        move_up = p_selected_index - p_first_shown_index;
        scroll_up = 0;
    }

    move_selection_up(move_up, scroll_up, scroll_options);
}

void
List::move_selection_up(unsigned move_up, unsigned scroll_up,
                        Scroll_Options scroll_options)
{
    // Make sure selection doesn't move past beginning
    if (move_up > p_selected_index)
    {
        move_up = p_selected_index;
        scroll_up = p_selected_index;
    }

    // Only invoke UI if there is something to do.
    if ((move_up != 0) || (scroll_up != 0))
    {
        p_ui->move_selection_up(move_up, scroll_up,
                                get_base_scroll_delay(scroll_options));
    }

    // These must be adjusted after the UI has been invoked because
    // the UI needs to know the old values in order to successfully
    // display the text for intermediate items.  For example, if
    // p_first_shown_index jumps from 10 to 5, the UI must scroll on
    // items 9 through 5.  If p_selected_index is set to 5 before
    // invoking the UI, then the UI does not know to start at 9.
    p_selected_index -= move_up;
    p_first_shown_index -= scroll_up;
}

unsigned
List::get_last_shown_index() const
{
    unsigned last_shown_index;
    
    // Calculate last_shown_index, clipping if necessary
    last_shown_index = p_first_shown_index + p_num_visible_items - 1;
    if (last_shown_index >= p_model->get_size())
        last_shown_index = p_model->get_size() - 1;

    return last_shown_index;
}

void
List::line_down(Scroll_Options scroll_options)
{
    unsigned move_down;
    unsigned scroll_down;

    if (! p_ui->is_unhidden())
        return;
    
    move_down = 1;
    if (p_selected_index == get_last_shown_index())
        scroll_down = 1;
    else
        scroll_down = 0;

    move_selection_down(move_down, scroll_down, scroll_options);
}

void
List::page_down(Scroll_Options scroll_options)
{
    unsigned move_down;
    unsigned scroll_down;
    unsigned last_shown_index;

    if (! p_ui->is_unhidden())
        return;
    
    last_shown_index = get_last_shown_index();
    if (p_selected_index == last_shown_index)
    {
        // At end, scroll down a page.
        move_down = p_num_visible_items - 1;
        scroll_down = p_num_visible_items - 1;
    }
    else
    {
        // Move to last shown.
        move_down = last_shown_index - p_selected_index;
        scroll_down = 0;
    }

    move_selection_down(move_down, scroll_down, scroll_options);
}

void
List::move_selection_down(unsigned move_down, unsigned scroll_down,
                          Scroll_Options scroll_options)
{
    unsigned size;

    // Make sure selection doesn't move past end
    size = p_model->get_size();
    if (size == 0)
    {
        move_down = 0;
        scroll_down = 0;
    }
    else if ((p_selected_index + move_down) > (size - 1))
    {
        move_down = size - 1 - p_selected_index;
        scroll_down = size - 1 - p_selected_index;
    }

    // Only invoke UI if there is something to do.
    if ((move_down != 0) || (scroll_down != 0))
    {
        p_ui->move_selection_down(move_down, scroll_down,
                                  get_base_scroll_delay(scroll_options));
    }

    // These must be adjusted after the UI has been invoked because
    // the UI needs to know the old values in order to successfully
    // display the text for intermediate items.  For example, if
    // p_first_shown_index jumps from 10 to 5, the UI must scroll on
    // items 9 through 5.  If p_selected_index is set to 5 before
    // invoking the UI, then the UI does not know to start at 9.
    p_selected_index += move_down;
    p_first_shown_index += scroll_down;
}

int
List::get_base_scroll_delay(Scroll_Options scroll_options) const
{
    int base_scroll_delay;

    switch (scroll_options)
    {
        case SMOOTH_SCROLL:
            base_scroll_delay = BASE_SCROLL_DELAY;
            break;

        case INSTANT_SCROLL:
        default:
            base_scroll_delay = 0;
            break;
    }
    return base_scroll_delay;
}

void
List::hide_split()
{
    p_ui->hide_split(1000);
}

void
List::hide_left()
{
    p_ui->hide_left(1000);
}

void
List::hide_right()
{
    p_ui->hide_right(1000);
}

void
List::hide_top(int milliseconds)
{
    p_ui->hide_top(milliseconds);
}

void
List::hide_split_center_selected(int milliseconds)
{
    p_ui->hide_split_center_selected(milliseconds);
}

void
List::unhide()
{
    p_ui->unhide(1000);
}

bool
List::update()
{
    if (p_list_changed)
    {
        p_ui->list_changed();
        p_list_changed = false;
    }
    
    return p_ui->update();
}

void
List::draw(BITMAP * bitmap)
{
    p_ui->draw(bitmap);
}


/************************************************************************
 *
 * List_UI
 *
 ************************************************************************/

List_UI::~List_UI()
{
}


/************************************************************************
 *
 * Abstract_List_UI
 *
 ************************************************************************/

Abstract_List_UI::Abstract_List_UI()
    : p_visible_items(), p_offscreen_items()
{
    p_container = new GEContainer();
    p_item_height = outline_text_height(UI_Manager::get_font());
}

Abstract_List_UI::~Abstract_List_UI()
{
    Label_List::iterator i;

    for (i = p_visible_items.begin(); i != p_visible_items.end(); i++)
    {
        delete (*i);
    }

    for (i = p_offscreen_items.begin(); i != p_offscreen_items.end(); i++)
    {
        delete (*i);
    }

    delete p_container;
}

unsigned
Abstract_List_UI::get_number_of_visible_items()
{
    return p_num_visible_items;
}

bool
Abstract_List_UI::update()
{
    bool container_updated;

    // The container must be updated before the state is changed as
    // items in the container may affect the state.
    container_updated = p_container->update();

    if ((p_state == HIDING_SPLIT) && are_all_items_stopped())
    {
        p_state = HIDDEN_SPLIT;
    }

    if ((p_state == HIDING_LEFT) && are_all_items_stopped())
    {
        p_state = HIDDEN_LEFT;
    }

    if ((p_state == HIDING_RIGHT) && are_all_items_stopped())
    {
        p_state = HIDDEN_RIGHT;
    }

    if ((p_state == HIDING_TOP) && are_all_items_stopped())
    {
        p_state = HIDDEN_TOP;
    }

    if ((p_state == UNHIDING) && are_all_items_stopped())
    {
        p_state = UNHIDDEN;
    }

    if (p_state == SCROLLING)
    {
        if (are_all_items_stopped())
            p_state = UNHIDDEN;
    }

    if (p_state == UNHIDDEN)
    {
        // list_updated = handle_input();
    }

    return container_updated;
}

void
Abstract_List_UI::draw(BITMAP * frame)
{
    // Set the clipping to be the same as the geometry of the parent
    // list so that offscreen items are really not displayed.
    set_clip(frame, p_x, p_y, p_x + p_width, p_y + p_height);
    p_container->draw(frame);
    set_clip(frame, 0, 0, SCREEN_W, SCREEN_H);
#if 0    
    rect(frame, p_x, p_y, p_x + p_width, p_y + p_height,
         makecol(255, 255, 255));
#endif
}

void
Abstract_List_UI::move_selection_up(unsigned move, unsigned scroll,
                                    int base_scroll_delay)
{
    unsigned    i;
    unsigned    first_shown_index;
    Engine *    engine;
    const List_Model * model;
    int         scroll_ms;
    int         screen_index;
    Label_List  moved_items;

    model = p_list->get_model();
    engine = UI_Manager::get_engine();
    scroll_ms = base_scroll_delay * ((scroll + 1)/ 2);
    first_shown_index = p_list->get_first_shown_index();

    unselect_selected();
    
    // First, perform the scroll
    for (i = 0; i < scroll; i++)
    {
        Label *         item;
        unsigned        next_index;

        // Remove last item, and stick it offscreen.
        item = p_visible_items.back();
        p_visible_items.pop_back();

        p_offscreen_items.push_front(item);
        moved_items.push_front(item);

        // Pull a selection from the offscreen list, setup the new text,
        // and move it to the top of the screen.
        item = p_offscreen_items.back();
        p_offscreen_items.pop_back();

        next_index      = first_shown_index - i - 1;
        screen_index    = 0 - int(i) - 1;

        item->set_contents(model->get_item_text_at(next_index),
                           model->get_item_icon_at(next_index));
        item->update();
        unhide_item(item, screen_index, 0);

        p_visible_items.push_front(item);

        engine->poll_background_task();
    }

    // Move selected item.  Must be done in a loop because iterators
    // do not implement the -= operator and thus must be iterated one
    // at a time.
    for (i = 0; i < move; i++)
    {
        p_selected_item--;
    }

    // Move positions
    if (scroll > 0)
    {
        Label_List::iterator  j;

        screen_index = 0;
        for (j = p_visible_items.begin(); j != p_visible_items.end(); j++)
        {
            unhide_item(*j, screen_index, scroll_ms);
            screen_index++;
        }

        // Put moved visible items past the bottom of the list
        for (j = moved_items.begin(); j != moved_items.end(); j++)
        {
            unhide_item(*j, screen_index, scroll_ms);
            screen_index++;
        }
        p_state = SCROLLING;
    }

    select_selected();
    
    return;
}

void
Abstract_List_UI::move_selection_down(unsigned move, unsigned scroll,
                                      int base_scroll_delay)
{
    unsigned    i;
    unsigned    selected_index;
    int         scroll_ms;
    int         screen_index;
    Label_List  moved_items;
    const List_Model * model;
    Engine *    engine;

    model = p_list->get_model();
    engine = UI_Manager::get_engine();
    scroll_ms = base_scroll_delay * ((scroll + 1)/ 2);
    selected_index = p_list->get_selected_index();

    unselect_selected();

    // First, perform the scroll
    for (i = 0; i < scroll; i++)
    {
        Label *        item;
        unsigned        next_index;

        // Remove first item, but stick it on the back of the offscreen
        // list.
        item    = p_visible_items.front();
        p_visible_items.pop_front();
        p_offscreen_items.push_back(item);
        moved_items.push_back(item);

        // Pull a selection off the front of the offscreen list, setup
        // the new text, and move it to the bottom.
        item    = p_offscreen_items.front();
        p_offscreen_items.pop_front();

        next_index      = selected_index + i + 1;
        screen_index    = p_num_visible_items + i;

        item->set_contents(model->get_item_text_at(next_index),
                           model->get_item_icon_at(next_index));
        item->update();
        unhide_item(item, screen_index, 0);

        p_visible_items.push_back(item);

        engine->poll_background_task();
    }

    // Move selected item.  Must be done in a loop because iterators
    // do not implement the += operator and thus must be iterated one
    // at a time.
    for (i = 0; i < move; i++)
    {
        p_selected_item++;
    }

    // Move positions
    if (scroll > 0)
    {
        Label_List::reverse_iterator  j;

        screen_index = p_num_visible_items - 1;
        for (j = p_visible_items.rbegin(); j != p_visible_items.rend(); j++)
        {
            unhide_item(*j, screen_index, scroll_ms);
            screen_index--;
        }

        for (j = moved_items.rbegin(); j != moved_items.rend(); j++)
        {
            unhide_item(*j, screen_index, scroll_ms);
            screen_index--;
        }

        p_state = SCROLLING;
    }

    select_selected();
    
    return;
}

void
Abstract_List_UI::list_changed()
{
    Label_List::iterator i;
    unsigned num_items;
    unsigned index;
    unsigned selected;
    const List_Model * model;

    num_items   = p_list->get_model()->get_size();
    index       = p_list->get_first_shown_index();
    selected    = p_list->get_selected_index();
    model       = p_list->get_model();
    p_justification = p_list->get_justification();

    for (i = p_visible_items.begin(); i != p_visible_items.end(); i++)
    {
        if (index >= num_items)
        {
            (*i)->set_contents("", 0);
            (*i)->set_background_color(makecol(255, 0, 255));
            (*i)->set_foreground_color(makecol(0, 0, 0));
            (*i)->set_outline_color(makecol(0, 0, 0), 0);
        }
        else
        {
            (*i)->set_contents(model->get_item_text_at(index),
                               model->get_item_icon_at(index));
        
            (*i)->set_background_color(makecol(255, 0, 255));

            if (index == selected)
            {
                p_selected_item = i;
                (*i)->set_foreground_color(p_list->get_selected_text_color());
                (*i)->set_outline_color(
                    p_list->get_selected_outline_color(), 800);
            }
            else
            {
                (*i)->set_foreground_color(
                    p_list->get_unselected_text_color());
                (*i)->set_outline_color(
                    p_list->get_unselected_outline_color(), 800);
            }
        }

        // Force un update of the label's contents
        (*i)->update();

        index++;
    }

    switch (p_state)
    {
        case HIDING_TOP:
        case HIDDEN_TOP:
            hide_top(0);
            break;
            
        case HIDING_SPLIT:
        case HIDDEN_SPLIT:
        default:
            hide_split(0);
            break;
    }
}

void
Abstract_List_UI::hide_offscreen_items()
{
    Label_List::iterator label;
    unsigned index;

    for (label = p_offscreen_items.begin(), index = 0;
         label != p_offscreen_items.end();
         label++, index++)
    {
        hide_offscreen_item(*label);
    }
}

void
Abstract_List_UI::unselect_selected()
{
    (*p_selected_item)->set_foreground_color(
        p_list->get_unselected_text_color());
    (*p_selected_item)->set_outline_color(
        p_list->get_unselected_outline_color(), 800);
}

void
Abstract_List_UI::select_selected()
{
    (*p_selected_item)->set_foreground_color(
        p_list->get_selected_text_color());
    (*p_selected_item)->set_outline_color(
        p_list->get_selected_outline_color(), 800);
}

void
Abstract_List_UI::hide_split(int milliseconds)
{
    Label_List::iterator label;
    unsigned index;

    for (label = p_visible_items.begin(), index = 0;
         label != p_visible_items.end();
         label++, index++)
    {
        if ((index % 2) == 0)
            hide_item_left(*label, index, milliseconds);
        else
            hide_item_right(*label, index, milliseconds);
    }

    hide_offscreen_items();
    p_state = HIDING_SPLIT;
}

void
Abstract_List_UI::hide_left(int milliseconds)
{
    Label_List::iterator label;
    unsigned index;

    for (label = p_visible_items.begin(), index = 0;
         label != p_visible_items.end();
         label++, index++)
    {
        hide_item_left(*label, index, milliseconds);
    }

    hide_offscreen_items();
    p_state = HIDING_LEFT;
}

void
Abstract_List_UI::hide_right(int milliseconds)
{
    Label_List::iterator label;
    unsigned index;

    for (label = p_visible_items.begin(), index = 0;
         label != p_visible_items.end();
         label++, index++)
    {
        hide_item_right(*label, index, milliseconds);
    }

    hide_offscreen_items();
    p_state = HIDING_RIGHT;
}

void
Abstract_List_UI::hide_top(int milliseconds)
{
    Label_List::iterator label;
    unsigned index;

    for (label = p_visible_items.begin(), index = 0;
         label != p_visible_items.end();
         label++, index++)
    {
        hide_item_top(*label, index, milliseconds);
    }

    hide_offscreen_items();
    p_state = HIDING_TOP;
}

void
Abstract_List_UI::hide_split_center_selected(int milliseconds)
{
    Label_List::iterator label;
    unsigned index;
    unsigned list_index;
    unsigned selected_index;

    list_index = p_list->get_first_shown_index();
    selected_index = p_list->get_selected_index();

    for (label = p_visible_items.begin(), index = 0;
         label != p_visible_items.end();
         label++, index++)
    {
        if (list_index == selected_index)
            center_item(*label, milliseconds);
        else if ((index % 2) == 0)
            hide_item_left(*label, index, milliseconds);
        else
            hide_item_right(*label, index, milliseconds);

        list_index++;
    }

    hide_offscreen_items();
    p_state = HIDING_SPLIT;
}

void
Abstract_List_UI::unhide(int milliseconds)
{
    Label_List::iterator label;
    int index;

    for (label = p_visible_items.begin(), index = 0;
         label != p_visible_items.end();
         label++, index++)
    {
        unhide_item(*label, index, milliseconds);
    }
    p_state = UNHIDING;
}

bool
Abstract_List_UI::is_unhidden() const
{
    return (p_state == UNHIDDEN);
}

bool
Abstract_List_UI::are_all_items_stopped() const
{
    Label_List::const_iterator i;
    bool        all_stopped;

    all_stopped = true;
    for (i = p_visible_items.begin(); i != p_visible_items.end(); i++)
    {
        if (! (*i)->is_stopped())
        {
            all_stopped = false;
            break;
        }
    }

    return all_stopped;
}
