#ifndef _GUI_COMP_H_
#define _GUI_COMP_H_

#include "engine.h"

/**
 * A collection of pseudo-GUI components for use in a game engine.
 * The components are designed in a similar fashion to Java Swing GUI
 * components with a pluggable look and feel.  The drawing of each
 * component is separated out into their own classes, called UI
 * classes.  The user never useses these UI classes directly and thus
 * can use the component regardless of which UI class is used.  This
 * creates a pluggable architeture where the UI class can be changed
 * without any user code modifications.
 *
 * @see GameEngine::Engine
 */
namespace GUI
{
    using GameEngine::Drawable;
    using GameEngine::Updatable;

    /**
     * A base interface for all GUI components.
     */
    class Component : public Drawable, public Updatable
    {
      public:
        /**
         * Virtual destructor for an interface.
         */
        virtual ~Component();

        /**
         * Updates the component for this cycle.
         *
         * @see GameEngine::Updatable::update()
         */
        virtual bool update() = 0;

        /**
         * Draws the component to the current animation frame.
         *
         * @see GameEngine::Drawable::draw()
         */
        virtual void draw(BITMAP * frame) = 0;
    };

    /**
     * All possible types of justification.
     */
    enum Justification
    {
        /** Justified against the left edge. */
        LEFT_JUSTIFIED,

        /** Justified against the center line. */
        CENTER_JUSTIFIED,

        /** Justified against the right edge. */
        RIGHT_JUSTIFIED,
    };

    /**
     * Used to select how to scroll the items.
     */
    enum Scroll_Options
    {
        /** Use smooth scrolling. */
        SMOOTH_SCROLL,
        /**
         * Do not use smooth scrolling.  The items move instantly
         * to their new locations.
         */
        INSTANT_SCROLL,
    };
};


#endif

/* Local Variables: */
/* mode: c++ */
/* End: */
