#ifndef _GUI_LABEL_H_
#define _GUI_LABEL_H_

#include "gui/comp.h"
#include "gui/util.h"

namespace GUI
{
    using GameEngine::Engine;

    class Label_UI;
    class Icon;

    /**
     * Short text with an optional icon.  The text is displayed with
     * an outline around it.  The text is not necessarily displayed
     * left to right on the screen.  The UI may display the text in
     * any fashion it desires.  This has importnat implications on the
     * coordinates used to specify the location and size of the label.
     * The text is displayed on the screen as a rectangle, but the
     * dimensions of the screen rectangle need not coorrespond the the
     * dimensions of the label.  For examle, the height of the label
     * may correspond to the height of the screen rectange, or the the
     * width of the screen rectangle depending on if the UI rotates
     * the text.  The location always corresponds to the upper left
     * corner of the screen where the label is to be displayed.
     *
     * When the label location is changed, it may be moved over time,
     * thus causing the label to visually move to the user.  The
     * outline color can also pulsate from the color given to black.
     * This is a nice blinking effect.
     *
     * The maximum width and height must be specified because video
     * memory may be used to speed up blitting.  It is best to
     * allocate video memory once and hold on to it rather than
     * constantly allocate and free it.  Thus, the video memory (if
     * used) is allocated all up front at object construction time and
     * only released at destruction time.
     *
     * Also, the height of the label text must not change for similar
     * reasons.  Note the height of the label text is not necessarily
     * the same as the height of the label, depending on the look and
     * feel.  This means the font must not change and the height of
     * the icon must never be taller than the height of the text.
     */
    class Label : public Component
    {
      public:
        /**
         * Constructs a new label specifying maximum screen
         * dimensions.  The label will most likely never have real
         * screen dimensions of this size as the label text limits the
         * logical height.  The logical height may correspond to
         * screen height or width depending on the look and feel.
         *
         * @param maximum_width Maximum width.
         * @param maximum_height Maximum height.
         */
        Label(int maximum_width, int maximum_height);

        /**
         * Destroys the label and any associated UIs and bitmaps.  The
         * icon is not destroyed.
         */
        ~Label();

        /**
         * Gets the maximum width in, screen pixels.
         *
         * @return Maximum width in, screen pixels.
         */
        int get_maximum_width() const;

        /**
         * Gets the maximum height, in screen pixels.
         *
         * @return Maximum height, in screen pixels.
         */
        int get_maximum_height() const;

        /**
         * Sets the upper left corner where the label will be drawn on
         * the screen.  If <tt>milliseconds</tt> is specified, then
         * the label will take that many milliseconds to reach the new
         * location.
         *
         * @param x X coordinate of upper left corner.
         * @param y Y coordinate of uper left corner.
         * @param milliseconds Number of millseconds to move to new
         * location.
         */
        void set_location(int x, int y, int milliseconds = 0);

        /**
         * Checks to see if the label is currently moving.
         *
         * @return True if the label is not moving.
         */
        bool is_stopped() const;

        /**
         * Gets the x coordinate of the upper left corner.
         *
         * @return X coordinate of upper left corner.
         */
        int get_x() const;

        /**
         * Gets the y coordinate of the upper left corner.
         *
         * @return Y coordinate of upper left corner.
         */
        int get_y() const;

        /**
         * Sets the actual size of the label.  This should only be
         * called by the UI.
         *
         * @param width Width of label, in screen pixels.
         * @param height Height of label, in screen pixels.
         */
        void set_size(int width, int height);

        /**
         * Gets the actual width of the label.
         *
         * @return Width of this label, in screen pixels.
         */
        int get_width() const;

        /**
         * Gets the actual height of the label.
         *
         * @return Height of this label, in screen pixels.
         */
        int get_height() const;

        /**
         * Sets the text and icon of this label.  If no icon is
         * desired, then set <tt>icon</tt> to <tt>null</tt>.  The icon
         * is not owned by the label and hence will not be deleted
         * when the label is destroyed.  It is up to the caller to
         * manager the icon.
         *
         * @param text Text of this label.
         * @param icon Icon to display, or <tt>null</tt> if no icon.
         */
        void set_contents(const string & text, const Icon * icon);

        /**
         * Gets the text of this label.
         *
         * @return Text of this label.
         */
        string get_text() const;

        /**
         * Gets the icon for this label.  If there is no label, then
         * null is returned.
         *
         * @return Icon of this label.
         */
        const Icon * get_icon() const;

        /**
         * Sets the foreground (text) color.
         *
         * @param foreground_color Foreground color.
         */
        void set_foreground_color(int foreground_color);

        /**
         * Gets the foreground (tex) color of this label.
         *
         * @return Foreground color.
         */
        int get_foreground_color();

        /**
         * Sets the background color.  If bright pink (255, 0, 255),
         * is used, the background is transparent since the label is
         * displayed using draw_sprite().  The default background
         * color is bright pink.
         */
        void set_background_color(int background_color);

        /**
         * Gets the background color of this label.
         *
         * @return Background color.
         */
        int get_background_color();

        /**
         * Sets the outline color of this label.  This color may
         * pulsate using the pulse period.  The pulse period is the
         * time it takes for the color to fade to black and then back
         * again.  If the pulse period is set to 0, then pulsating is
         * disabled and the color is shown as solid.
         *
         * @param outline_color Outline color.
         * @param pulse_period The period, in milliseconds, of the the
         * pulse to black or 0 to disable pulsing.
         */
        void set_outline_color(int outline_color, int pulse_period);

        /**
         * Gets the outline color of this label.
         *
         * @return Outline color.
         */
        int get_outline_color();

        /**
         * Gets the font used for the text.
         *
         * @return Font used for the text.
         */
        FONT * get_font() const;

        /**
         * Updates the location and outline color.
         *
         * @return True if anything actually changed.
         */
        virtual bool update();

        /**
         * Draws the label to the frame.
         *
         * @param frame The current animation frame.
         */
        virtual void draw(BITMAP * frame);

      private:
        /** The UI delegate for this label. */
        Label_UI * p_ui;

        /** The current location of the upper left corner. */
        Movable_Point p_location;

        /** The width of this label as it appers on the screen. */
        int p_width;

        /** The height of this label as it appears on the screen. */
        int p_height;

        /** The absolute maximum width the label can possibly be. */
        int p_max_width;

        /** The absolute maximum height the label can possibley be. */
        int p_max_height;

        /** The color of the foreground (text). */
        int p_foreground_color;

        /** The color of the background. */
        int p_background_color;

        /** The outline color. */
        Pulsating_Color p_outline_color;

        /** The current font. */
        FONT * p_font;

        /** The text for this label. */
        string p_text;

        /** The icone for this label. */
        const Icon * p_icon;

        /**
         * Set to true if any of the parameters other than location
         * have been changed.  This is used to only update the UI once
         * during the update() method instead of every time a paramter
         * is changed.
         */
        bool p_updated;
    };

    /**
     * An interface for the UI delegate for a Label.  It is
     * responsible for drawing the label to the screen.
     */
    class Label_UI
    {
      public:
        /**
         * Virtual destructor.
         */
        virtual ~Label_UI();

        /**
         * Called when the UI is first associated with a Label.  It
         * should only be called once at setup as a video bitmap may
         * be used for performance reasons and reallocating video
         * bitmaps is costly.
         */
        virtual void install_ui(Label * label) = 0;

        /**
         * Called when the label has been changed.  This should redraw
         * any local bitmap of the label.
         */
        virtual void label_updated() = 0;

        /**
         * Called when the label needs to be drawn to the screen.
         *
         * @param bitmap Bitmap to draw on.
         */
        virtual void draw(BITMAP * bitmap) = 0;

    };

    /**
     * A base class for concrete Label_UI implementations.  Two
     * bitmaps are provided for subclasses.  The main bitmap is
     * created at install time and is the maximum size of the label.
     * This may be a video bitmap, which is why it is created at
     * install time and never modified.  This bitmap must be allocated
     * by the subclass since the dimensions are not known.  The label
     * bitmap is the the bitmap for the label is it will be displayed.
     * It is always a sub-bitmap of the main bitmap.  Some helper
     * methods are also provided to help simplify subclasses.
     */
    class Abstract_Label_UI : public Label_UI
    {
      public:
        /**
         * Sets both bitmaps to null.
         */
        Abstract_Label_UI();

        /**
         * Destroys bitmaps.
         *
         * @see destroy_bitmap()
         */
        virtual ~Abstract_Label_UI();
        
        /**
         * Draws the <tt>p_label_bitmap</tt> to the bitmap.
         */
        virtual void draw(BITMAP * bitmap);

      protected:
        /**
         * Used to select which bitmap should be used in
         * draw_horizontal().
         */
        enum Destination_Bitmap
        {
            /** Use a sub-bitmap of the p_main_bitmap. */
            SUB_BITMAP,
            /** Create a new, temporary memory bitmap. */
            TEMPORARY_BITMAP
        };

        /**
         * Destroys <tt>p_main_bitmap</tt> and
         * <tt>p_label_bitmap</tt>.
         */
        void destroy_bitmaps();

        /**
         * Truncates text until it fits within the maximum width.
         *
         * @param text May be truncated to fit within the maximum width.
         * @param font The font to use when calculating text widths.
         * @param maximum_width The constraining width.
         * @return The actual width, in pixels, of the new text.
         */
        int truncate_text(string & text, FONT * font, int maximum_width);

        /**
         * Draws the label horizontal (left to right).  It is either
         * drawn on a sub-bitmap of <tt>p_main_label</tt> or a new
         * memory bitmap.  The bitmap width will never be larger than
         * <tt>maximum_width</tt>, but will always have the height as
         * specified.
         *
         * @param destination_bitmap Which bitmap to draw on.
         * @param maximum_width The constrainig width.
         * @param height The height of the bitmap.
         * @return The bitmap that was drawn on.
         */
        BITMAP * draw_horizontal(Destination_Bitmap destination_bitmap,
                                 int maximum_width, int height);

        /** The parent Label. */
        Label * p_label;

        /** The main bitmap. */
        BITMAP * p_main_bitmap;
        
        /**
         * A sub-bitmap of the main bitmap, but always contains the
         * current label.
         */
        BITMAP * p_label_bitmap;
    };
};

#endif

/* Local Variables: */
/* mode: c++ */
/* End: */
