#ifndef _GUI_MENU_H_
#define _GUI_MENU_H_

#include "gui/comp.h"
#include "gui/list.h"

namespace GUI
{
    class Icon;
    class Label;
    class List;
    class List_Model;
    class Menu_UI;
    class Menu_Title_List_Adapter;
    class Separator;

    /**
     * A menu is an aggregate two labels and a list.  The labels are
     * for a title and sub-title and the list is for the menu items.
     * The labels appear above the list.
     */
    class Menu : public Component
    {
      public:
        /**
         * Contsruct a new menu with an empty title, subtitle, and list.
         */
        Menu(int x, int y, int width, int height);

        /**
         * Destructor;
         */
        ~Menu();

        /**
         * Gets the x coordinate of the upper left corner.
         *
         * @return X coordinate of the upper left corner.
         */
        int get_x() const;

        /**
         * Gets the y coordinate of the upper left corner.
         *
         * @return Y coordinate of the upper left corner.
         */
        int get_y() const;
        
        /**
         * Gets the width of this list.
         *
         * @return The width of this list.
         */
        int get_width() const;

        /**
         * Gets the height of this list.
         *
         * @return The height of this list.
         */
        int get_height() const;
        
        /**
         * Sets the title for this menu.
         *
         * @param title Text of the title.
         */
        void set_title(const string & title);

        /**
         * Gets the title of this menu.
         *
         * @return The title.
         */
        string get_title() const;

        /**
         * Sets the subtitle of this menu.
         *
         * @param subtitle Text of the subtitle.
         */
        void set_subtitle(const string & subtitle);

        /**
         * Gets the subtitle of this menu.
         *
         * @return The subtitle.
         */
        string get_subtitle() const;

        /**
         * Sets the model for the embedded list.
         *
         * @param model The new model.
         */
        void set_list_model(const List_Model * model);
        
        /**
         * Gets the model for the embedded list.
         *
         * @return The model for this list.
         */
        const List_Model * get_list_model() const;
        
        /**
         * Sets the justification for all items in this menu.
         * The default is centered justification.
         *
         * @param justification The justification of the items.
         */
        void set_justification(Justification justification);

        /**
         * Gets the justification for all items in this menu.
         *
         * @return The justificaiton of the items.
         */
        Justification get_justification() const;
        
        
        /**
         * Sets the position of the selection.  The first shown index
         * is the first item to be visible in the list.  This allows
         * the selection to be set to any item displayed.
         *
         * @param first_shown_index Index of the first displayed item.
         * @param selected_index Index of selected item.
         */
        void set_selection_position(unsigned first_shown_index,
                                    unsigned selected_index);

        /**
         * Gets the index of the first item visible in this list.
         *
         * @return Index of the first shown item.
         */
        unsigned get_first_shown_index() const;
        
        /**
         * Gets the index of the selected item.
         *
         * @return Index of the selected item.
         */
        unsigned get_selected_index() const;
        
        /**
         * Sets the colors to use for the selected item.
         *
         * @param text Text color.
         * @param outline Outline color.
         */
        void set_selected_colors(int text, int outline);

        /**
         * Gets the text color to be used for the selected item.
         *
         * @return Text color for the selected item.
         */
        int get_selected_text_color() const;

        /**
         * Gets the outline color to be used for the selected item.
         *
         * @return Outline color for the selected item.
         */
        int get_selected_outline_color() const;
        
        /**
         * Sets the color to use for unselected items.
         *
         * @param text Text color.
         * @param outline Outline color.
         */
        void set_unselected_colors(int text, int outline);

        /**
         * Gets the text color to be used for unselected items.
         *
         * @return Text color for unselected items.
         */
        int get_unselected_text_color() const;

        /**
         * Gets the outline color to be used for unselected items.
         *
         * @return Outline color for unselected items.
         */
        int get_unselected_outline_color() const;
        
        /**
         * Move the selection up one line.  If the selection is at the
         * top of the displayed items, all items are scrolled down by
         * one.
         *
         * @param scroll_options Options to use, if scrolled.
         */
        void line_up(Scroll_Options scroll_options);

        /**
         * Move the selection down one line.  If the selection is at
         * the bottom of the displayed items, all items are scrolled
         * up by one.
         *
         * @param scroll_options Options to use, if scrolled.
         */
        void line_down(Scroll_Options scroll_options);

        /**
         * Move the selection up one page.
         *
         * @param scroll_options Options to use, if scrolled.
         */
        void page_up(Scroll_Options scroll_options);

        /**
         * Move the selection down one page.
         *
         * @param scroll_options Options to use, if scrolled.
         */
        void page_down(Scroll_Options scroll_options);

        /**
         * Hide all items with half going to the left and half going
         * to the right.
         */
        void hide_split();

        /**
         * Hide all items to the left.
         */
        void hide_left();

        /**
         * Hide all items to the right.
         */
        void hide_right();
        
        /**
         * The same as hide_split() but the selected item stays
         * displayed and is centered within the list geometry.
         *
         * @param milliseconds.
         */
        void hide_split_center_selected(int milliseconds);

        /**
         * Unhides all items.
         */
        void unhide();

        /**
         * Updates the position of all items.
         */
        virtual bool update();

        /**
         * Draws all items at their current location.
         *
         * @param frame Active screen frame.
         */
        virtual void draw(BITMAP * frame);

      private:
        /** The UI for this list. */
        Menu_UI * p_ui;
        
        /**
         * Set to true if the list has changed.  This is used to
         * update the UI once during update() instead of for every
         * parameter that has changed.
         */
        bool p_has_changed;
        
        /** The x-coordinate of the upper left corner of this menu. */
        int p_x;

        /** The y-coordinate of the upper left corner of this menu. */
        int p_y;

        /** The width of this menu. */
        int p_width;

        /** The height of this menu. */
        int p_height;
        
        /** The text for the title. */
        string p_title;

        /** The text for the subtitle. */
        string p_subtitle;

        /** The text color for the selected item. */
        int p_selected_text_color;

        /** The outline color for the selected item. */
        int p_selected_outline_color;

        /** The text color for the unselected items. */
        int p_unselected_text_color;

        /** The outline color for the unselected items. */
        int p_unselected_outline_color;

        /** Justification of items in title and menu lists. */
        Justification p_justification;
        
        /**
         * A local pointer to the list used to implement the title in
         * the UI.
         */
        List * p_title_list;

        /**
         * A local pointer to the list used to implement the menu
         * items in the UI.
         */
        List * p_menu_list;
    };

    /**
     * The interface that draws the menu.  The title and subtitle
     * labels are aggregated into a list, even though most of the
     * features of the list should not be used.  It is essentially a
     * fixed list that never scrolls and has no selected item.  The
     * selected item should be bypassed by setting the selected and
     * unselected colors to be the same.  Some of this implementation
     * is exposed to the Menu class because it is too require many
     * delegating methods in the UI interface.  This delegation does
     * not seem worth the encapsulation since Menu and Menu_UI are
     * fairly tightly coupled anyways.
     */
    class Menu_UI : public Updatable, public Drawable
    {
      public:
        /**
         * Virtual destructor.
         */
        virtual ~Menu_UI();

        /**
         * Called when a menu is attached to this UI.
         *
         * @param menu The parent menu.
         */
        virtual void install_menu(Menu * menu) = 0;

        /**
         * Notifies the UI that the parent menu has changed.
         */
        virtual void menu_changed() = 0;

        /**
         * Gets the aggregate list used for the titles.
         *
         * @return Title list.
         */
        virtual List * get_title_list() const = 0;

        /**
         * Gets the aggregate list used for the menu items.
         *
         * @return Menu item list.
         */
        virtual List * get_menu_list() const = 0;

        /**
         * Updates the position of any internal components.
         *
         * @return True if any of the components have changed.
         */
        virtual bool update() = 0;

        /**
         * Draws all internal components at their current location.
         *
         * @param frame Active screen frame.
         */
        virtual void draw(BITMAP * frame) = 0;
    };

    /**
     * An base implementation using two lists to represent a menu.
     */
    class Abstract_Menu_UI : public Menu_UI
    {
      public:
        /**
         * Constructor.
         */
        Abstract_Menu_UI();

        /**
         * Virtual destructor.
         */
        virtual ~Abstract_Menu_UI();

        /**
         * Called when a menu is attached to this UI.  Subclasses
         * should create and place the title and menu lists.
         *
         * @param menu The parent menu.
         */
        virtual void install_menu(Menu * menu) = 0;

        virtual void menu_changed();

        virtual List * get_title_list() const;

        virtual List * get_menu_list() const;

        virtual bool update();

        virtual void draw(BITMAP * frame);

      protected:
        /** The height of the separator. */
        static const int SEPARATOR_HEIGHT       = 6;

        /** The height of each item. */
        int p_item_height;

        /** The parent menu. */
        Menu * p_menu;

        /**
         * The list model used to provide the fixed list for the title
         * and subtitle.
         */
        Menu_Title_List_Adapter * p_title_model;

        /** The list for the titles. */
        List * p_title_list;

        /** The list for the menu items.*/
        List * p_menu_list;

        Separator * p_separator;
    };

    /**
     * Adapts the menu titles to a list model.  It is a fixed list
     * that has only 2 items.  The first item is the menu title and
     * the second item is the menu subtitle.  No icons are supported.
     */
    class Menu_Title_List_Adapter : public List_Model
    {
      public:
        Menu_Title_List_Adapter(Menu * menu);

        /**
         * Returns the title for index 0 and the subtitle for index 1.
         *
         * @param index Chooses title or subitle.
         * @return Title or subtitle.
         */
        virtual string get_item_text_at(unsigned index) const;

        /**
         * Always returns null.
         *
         * @return Always returns null (0).
         */
        virtual Icon * get_item_icon_at(unsigned index) const;

        /**
         * Always returns 2, one item for the title and one for the
         * subtitle.
         *
         * @return Always returns 2.
         */
        virtual unsigned get_size() const;

      private:
        /** The menu to use for the title and subtitle. */
        Menu * p_menu;
    };

};

#endif

/* Local Variables: */
/* mode: c++ */
/* End: */
