#ifndef _GUI_UIMGR_H
#define _GUI_UIMGR_H

#include "allegro.h"
#include "engine.h"

namespace GUI
{
    using GameEngine::Engine;

    class Label_UI;
    class List_UI;
    class Menu_UI;
    class Separator_UI;

    /**
     * The interface of a factory for GUI components.
     */
    class UI_Factory
    {
      public:
        /**
         * Virtual destructor for interfaces.
         */
        virtual ~UI_Factory();

        /**
         * Creates a new label UI.
         *
         * @return A new label UI.
         * @see GUI::Label
         */
        virtual Label_UI * create_label_ui() = 0;

        /**
         * Creates a new list UI.
         *
         * @return A new list UI.
         * @see GUI::List
         */
        virtual List_UI * create_list_ui() = 0;

        /**
         * Creates a new menu UI.
         *
         * @return A new menu UI.
         * @see GUI::Menu
         */
        virtual Menu_UI * create_menu_ui() = 0;

        /**
         * Creates a new separatro UI.
         *
         * @return A new separator UI.
         * @see GUI::Separator
         */
        virtual Separator_UI * create_separator_ui() = 0;
    };

    /**
     * A collection of static methods for the current UI delegates and
     * other objects that do not change.
     */
    class UI_Manager
    {
      public:
        /**
         * Sets the font that all future components will use.
         *
         * @param font New font.
         */
        static void set_font(FONT * font);
        
        /**
         * Gets the current font.
         *
         * @return The current font.
         */
        static FONT * get_font();
        
        /**
         * Sets the current game engine.
         *
         * @param engine New game engine.
         */
        static void set_engine(Engine * engine);

        /**
         * Gets the current game engine.
         *
         * @return Current game engine.
         */
        static Engine * get_engine();

        /**
         * Creates a new default label UI.
         *
         * @return A new label UI.
         */
        static Label_UI * get_label_ui();

        /**
         * Creates a new default list UI.
         *
         * @return A new list UI.
         */
        static List_UI * get_list_ui();
        
        /**
         * Creates a new default menu UI.
         *
         * @return A new menu UI.
         */
        static Menu_UI * get_menu_ui();

        /**
         * Creates a new default separator UI.
         */
        static Separator_UI * get_separator_ui();

        /**
         * Installs the current UI factory that creates all future UI
         * objects.
         *
         * @param factory New UI factory.
         */
        static void install_ui_factory(UI_Factory * factory);

      private:
        /** The current font. */
        static FONT * s_font;

        /** The current game engine. */
        static Engine * s_engine;

        /** The current UI factory. */
        static UI_Factory * s_ui_factory;
    };
};

#endif

/* Local Variables: */
/* mode: c++ */
/* End: */
