#ifndef _ENGINE_H_
#define _ENGINE_H_

#include <string>
#include <map>
#include <vector>

namespace MameScan
{
    typedef enum { GOOD, PRELIMINARY } GameStatus;

    class Exception
    {
      public:
        Exception(const string & error);
        virtual ~Exception();
        
        virtual const string & getError();

      protected:
        string      mError;
    };

    class GameInfo
    {
      public:
        GameInfo();

        char        name[10];
        char        description[80];
        char        romOf[10];
        char        cloneOf[10];
        GameStatus  status;

        bool            isClone();
    };

    class Constraint
    {
      public:
        typedef enum { NO_CONSTRAINT, EXCLUDE, INCLUDE } Result;

        virtual ~Constraint();
        
        virtual Result  matchConstraint(const string & value) = 0;
        virtual string  getConstraint() = 0;

      protected:
        string  mPattern;
    };

    class ExcludeConstraint : public Constraint
    {
      public:
        ExcludeConstraint(const string & pattern);

        Result  matchConstraint(const string & value);
        string  getConstraint();
    };

    class IncludeConstraint : public Constraint
    {
      public:
        IncludeConstraint(const string & pattern);

        Result  matchConstraint(const string & value);
        string  getConstraint();
    };

    class GameList
    {
      public:
        GameList();
        ~GameList();

        void    parseInfoFile(const string & infoFileName) throw (Exception);
        GameInfo *    lookupByName(string & name);
        vector<GameInfo *>  lookupByRomOf(string & romOf);

      private:
        typedef enum {IN_GAME, NOT_IN_GAME} State;
        typedef map<string, GameInfo *> NameList;
        typedef multimap<string, GameInfo *> RomOfList;

        static const char * GAME_START_MATCH;
        static const char * GAME_END_MATCH;
        static const char * NAME_MATCH;
        static const char * DESCRIPTION_MATCH;
        static const char * ROM_OF_MATCH;
        static const char * CLONE_OF_MATCH;
        static const char * STATUS_MATCH;
        static const char * STATUS_GOOD_MATCH;

        bool    matchGameStart(const char * line);
        bool    matchGameEnd(const char * line);
        bool    matchName(const char * line, char * match, size_t max_len);
        bool    matchDescription(const char * line, char * match,
                                 size_t max_len);
        bool    matchCloneOf(const char * line, char * match, size_t max_len);
        bool    matchRomOf(const char * line, char * match, size_t max_len);
        bool    matchStatus(const char * line, GameStatus * status);

        int         mGameStartSize;
        int         mGameEndSize;
        int         mNameSize;
        int         mDescriptionSize;
        int         mRomOfSize;
        int         mCloneOfSize;
        int         mStatusSize;
    
        NameList    mNameList;
        RomOfList   mRomOfList;
        State       mState;
    };
    

    class ConfigGenerator
    {
      public:
        ConfigGenerator(const string & infoFileName,
                        const string & configFileBaseName,
                        const string & romDirectory);
        ~ConfigGenerator();

        void    generateConfig() throw (Exception);

      private:
        void    parseConstraintFile(const string & constraintFileName);
        void    findGames() throw (Exception);
        void    outputConfigFiles() throw (Exception);
        void    printGameInfo(GameInfo * info);

        Constraint::Result  matchConstraints(const string & game);

        string          mInfoFileName;
        string          mConfigFileBaseName;
        string          mRomDirectory;
        GameList        mGameList;
        vector<GameInfo *>      mMatchedGames;
        vector<Constraint *>    mConstraints;
    };

    class Scanner
    {
      public:
        Scanner(const string & mameExecutable, const string & mameRomDirectory,
                const string & configurationName) throw (Exception);
        Scanner(const string & configFile) throw (Exception);

        void    scan() throw (Exception);
        
        static void cleanupTemporaryFiles();

      private:
        static const char * INFO_FILE;
        static const char * MAME_EXECUTABLE_PROPERTY;
        static const char * MAME_ROM_DIRECTORY_PROPERTY;
        static const char * BASE_CONFIGURATION_PATH_PROPERTY;
        
        bool    parseNameValuePair(const char * line, string & name,
                                   string & value);

        string          mMameExecutable;
        string          mMameRomDirectory;
        string          mConfigurationName;
    };

    extern const char * program_name;
}

void    splitExtension(char * path, char ** extension);

#define MIN(_A_, _B_)   ((_A_ < _B_) ? _A_ : _B_)

#endif

/* Local Variables: */
/* mode: c++ */
/* End: */
