#include <stdio.h>
#include <errno.h>

#include "mamescan.h"
#include "options.h"

using namespace MameScan;

#define _S(_A_, _B_) \
const char * GameList:: _A_ = _B_

_S(GAME_START_MATCH,    "game (");
_S(GAME_END_MATCH,      ")");
_S(NAME_MATCH,          "\tname");
_S(DESCRIPTION_MATCH,   "\tdescription");
_S(ROM_OF_MATCH,        "\tromof");
_S(CLONE_OF_MATCH,      "\tcloneof");
_S(STATUS_MATCH,        "\tdriver ( status");
_S(STATUS_GOOD_MATCH,   "good");

#undef _S


GameList::GameList()
    : mNameList(), mRomOfList()
{
    mState              = NOT_IN_GAME;
    mGameStartSize      = strlen(GAME_START_MATCH);
    mGameEndSize        = strlen(GAME_END_MATCH);
    mNameSize           = strlen(NAME_MATCH);
    mDescriptionSize    = strlen(DESCRIPTION_MATCH);
    mRomOfSize          = strlen(ROM_OF_MATCH);
    mCloneOfSize        = strlen(CLONE_OF_MATCH);
    mStatusSize         = strlen(STATUS_MATCH);
}


GameList::~GameList()
{
    NameList::iterator  i;

    for (i = mNameList.begin(); i != mNameList.end(); i++)
    {
        GameInfo *  info;

        info    = (*i).second;
        // printf("\r%-8s", info->name);
        delete  info;
    }
    printf("\n");
}

void
GameList::parseInfoFile(const string & infoFile) throw (Exception)
{
    char    line[1000];
    State   nextState;
    int     romCount;
    string  lineString;
    FILE *  file;
    
    const char *    romOf;
    const char *    cloneOf;
    GameStatus      status;
    GameInfo *      info;
    Options *       options;

    options     = Options::instance();
    file        = fopen(infoFile.c_str(), "r");
    romCount    = 0;
    info        = 0;

    if (file == NULL)
    {
        throw Exception(infoFile + ": " + strerror(errno));
    }
        
    while (1)
    {
        if (fgets(line, sizeof(line), file) == NULL)
            break;

        // Start off with out changing state
        nextState   = mState;
        
        if (mState == NOT_IN_GAME)
        {
            if (matchGameStart(line))
            {
                info    = new GameInfo();

                // Clear out values;
                romOf   = "";
                cloneOf = "";
                status  = PRELIMINARY;
                
                romCount++;
                nextState   = IN_GAME;
            }
        }
        else if (mState == IN_GAME)
        {
            if (matchGameEnd(line))
            {
                mNameList[string(info->name)] = info;
                if (strcmp(info->romOf, "") != 0)
                    mRomOfList.insert(make_pair(string(info->romOf), info));
                
                if (options->dump_info)
                    printf("\n");
                nextState   = NOT_IN_GAME;
            }

            if (matchName(line, info->name, sizeof(info->name)))
            {
                if (options->dump_info)
                    printf("Name: >%s<\n", info->name);
                else
                {
                    if (options->debug)
                        printf("\rROM #%4d, Name: %-8s", romCount, info->name);
                }
            }

            if (matchDescription(line, info->description,
                                 sizeof(info->description)))
            {
                if (options->dump_info)
                    printf("Description: >%s<\n", info->description);
            }


            if (matchRomOf(line, info->romOf, sizeof(info->romOf)))
            {
                if (options->dump_info)
                    printf("Rom Of >%s<\n", info->romOf);
            }

            if (matchCloneOf(line, info->cloneOf,
                             sizeof(info->cloneOf)))
            {
                if (options->dump_info)
                    printf("Clone Of >%s<\n", info->cloneOf);
            }


            if (matchStatus(line, &info->status))
            {
                if (options->dump_info)
                {
                    printf("Status >%s<\n", (info->status == GOOD ? "good" :
                                             "prelim"));
                }
            }
        }

        mState  = nextState;
    }

    fclose(file);
    return;
}

bool
GameList::matchGameStart(const char * line)
{
    if (strncmp(line, GAME_START_MATCH, mGameStartSize) == 0)
        return true;
    else
        return false;
}


bool
GameList::matchGameEnd(const char * line)
{
    if (strncmp(line, GAME_END_MATCH, mGameEndSize) == 0)
        return true;
    else
        return false;
}

bool
GameList::matchName(const char * line, char * match, size_t max_len)
{
    const char *    first;
    const char *    last;
    size_t          num_chars;

    if (strncmp(line, NAME_MATCH, mNameSize) == 0)
    {
        first = line + mNameSize;
        
        while (isspace(*first))
            first++;

        last = first;

        while (! isspace(*last))
            last++;

        num_chars   = MIN(size_t(last - first), max_len);
        strncpy(match, first, num_chars);
        match[num_chars] = '\0';
        return true;
    }
    else
        return false;
}

bool
GameList::matchDescription(const char * line, char * match, size_t max_len)
{
    const char *    first;
    const char *    last;
    size_t          num_chars;

    if (strncmp(line, DESCRIPTION_MATCH, mDescriptionSize) == 0)
    {
        first = line + mDescriptionSize;
        
        while ((*first) != '"')
            first++;
        first++;

        last = first;

        while (*last != '"')
            last++;

        num_chars   = MIN(size_t(last - first), max_len);
        strncpy(match, first, num_chars);
        match[num_chars] = '\0';
        return true;
    }
    else
        return false;
}

bool
GameList::matchRomOf(const char * line, char * match, size_t max_len)
{
    const char *    first;
    const char *    last;
    size_t          num_chars;

    if (strncmp(line, ROM_OF_MATCH, mRomOfSize) == 0)
    {
        first = line + mRomOfSize;
        
        while (isspace(*first))
            first++;

        last = first;

        while (! isspace(*last))
            last++;

        num_chars   = MIN(size_t(last - first), max_len);
        strncpy(match, first, num_chars);
        match[num_chars] = '\0';
        return true;
    }
    else
        return false;
    
}

bool
GameList::matchCloneOf(const char * line, char * match, size_t max_len)
{
    const char *    first;
    const char *    last;
    size_t          num_chars;

    if (strncmp(line, CLONE_OF_MATCH, mCloneOfSize) == 0)
    {
        first = line + mCloneOfSize;
        
        while (isspace(*first))
            first++;

        last = first;

        while (! isspace(*last))
            last++;

        num_chars   = MIN(size_t(last - first), max_len);
        strncpy(match, first, num_chars);
        match[num_chars] = '\0';
        return true;
    }
    else
        return false;
    
}

bool
GameList::matchStatus(const char * line, GameStatus * status)
{
    const char * first;
    const char * last;
    char    statusString[30];
    int     num_chars;

    if (strncmp(line, STATUS_MATCH, mStatusSize) == 0)
    {
        first = line + mStatusSize;
        
        while (isspace(*first))
            first++;

        last = first;

        while (! isspace(*last))
            last++;

        num_chars   = MIN(size_t(last - first), sizeof(statusString));
        strncpy(statusString, first, num_chars);
        statusString[num_chars] = '\0';

        if (strcmp(statusString, STATUS_GOOD_MATCH) == 0)
            *status = GOOD;
        else
            *status = PRELIMINARY;

        return true;
    }
    else
        return false;
    
}


GameInfo *
GameList::lookupByName(string & name)
{
    return mNameList[name];
}

vector<GameInfo *>
GameList::lookupByRomOf(string & name)
{
    vector<GameInfo *>  games;

    typedef RomOfList::const_iterator   I;

    pair<I, I> range = mRomOfList.equal_range(name);

    for (I i = range.first; i != range.second; i++)
    {
        games.push_back(i->second);
    }

    return games;
}
