#include <stdio.h>
#include "mamescan.h"
#include "options.h"

using namespace MameScan;

#define _S(_A_, _B_) \
const char * Scanner:: _A_ = _B_

_S(INFO_FILE,                           "_info.tmp");
_S(MAME_EXECUTABLE_PROPERTY,            "mame_executable");
_S(MAME_ROM_DIRECTORY_PROPERTY,         "mame_rom_directory");
_S(BASE_CONFIGURATION_PATH_PROPERTY,    "base_configuration_path");

#undef _S

Scanner::Scanner(const string & mameExecutable,
                 const string & mameRomDirectory,
                 const string & configurationName) throw (Exception)
{
    mMameExecutable     = mameExecutable;
    mMameRomDirectory   = mameRomDirectory;
    mConfigurationName  = configurationName;
}

Scanner::Scanner(const string & configFileName) throw (Exception)
{
    FILE *  configFile;
    char    line[1000];
    string  name;
    string  value;
    bool    foundExecutable;
    bool    foundRomDirectory;
    bool    foundConfigurationName;

    configFile  = fopen(configFileName.c_str(), "r");
    if (configFile == NULL)
    {
        throw Exception(configFileName +
                        ": Unable to open configuration file");
    }

    foundExecutable         = false;
    foundRomDirectory       = false;
    foundConfigurationName  = false;

    while (1)
    {
        if (fgets(line, sizeof(line), configFile) == NULL)
            break;

        if (parseNameValuePair(line, name, value))
        {
            if (name == MAME_EXECUTABLE_PROPERTY)
            {
                foundExecutable = true;
                mMameExecutable = value;
            }
            
            if (name == MAME_ROM_DIRECTORY_PROPERTY)
            {
                foundRomDirectory   = true;
                mMameRomDirectory   = value;
            }
            
            if (name == BASE_CONFIGURATION_PATH_PROPERTY)
            {
                foundConfigurationName  = true;
                mConfigurationName      = value;
            }
        }
    }

    fclose(configFile);

    if (! foundExecutable)
    {
        throw Exception(string(MAME_EXECUTABLE_PROPERTY) +
                        ": property not found");
    }

    if (! foundRomDirectory)
    {
        throw Exception(string(MAME_ROM_DIRECTORY_PROPERTY) +
                        ": property not found");
    }

    if (! foundConfigurationName)
    {
        throw Exception(string(BASE_CONFIGURATION_PATH_PROPERTY) +
                        ": property not found");
    }
}

bool
Scanner::parseNameValuePair(const char * line, string & name, string & value)
{
    const char *    nameFirst;
    const char *    nameLast;
    const char *    valueFirst;
    const char *    valueLast;
    const char *    eolChecker;
    char            nameString[100];
    char            valueString[100];
    int             length;
    
    atexit(Scanner::cleanupTemporaryFiles);

    // Here's a what we're trying to parse:
    //   name=value
    // There can be white space before name and before and after '='

    nameFirst   = line;
    while(isspace(*nameFirst))
        nameFirst++;

    if (*nameFirst == '\0')
        return false;
    
    nameLast    = nameFirst;
    while((! isspace(*nameLast)) && (*nameLast != '='))
        nameLast++;

    if (nameFirst == nameLast)
        return false;

    length  = MIN(size_t(nameLast - nameFirst), sizeof(nameString) - 1);
    strncpy(nameString, nameFirst, length);
    nameString[length] = '\0';

    valueFirst  = nameLast;
    while (isspace(*valueFirst))
        valueFirst++;

    // Ok, no name/value separator found, ditch out
    if (*valueFirst != '=')
        return false;
    
    // Skip over '='
    valueFirst++;

    while(isspace(*valueFirst))
        valueFirst++;

    if (*valueFirst == '\0')
        return false;
    
    valueLast = valueFirst;
    while (!isspace(*valueLast))
        valueLast++;

    if (valueFirst == valueLast)
        return false;
    
    length  = MIN(size_t(valueLast - valueFirst), sizeof(valueString) - 1);
    strncpy(valueString, valueFirst, length);
    valueString[length] = '\0';

    // Make sure there is only whitespace on the end
    eolChecker = valueLast;

    while (isspace(*eolChecker))
        eolChecker++;

    // Oops there's more text!
    if (*eolChecker != '\0')
        return false;

    // Copy into output parameters
    name    = string(nameString);
    value   = string(valueString);

    return true;
}

void
Scanner::scan() throw (Exception)
{
    int     rc;
    string  mameCommand;

    Options *           options;
    ConfigGenerator*    configGenerator;

    options     = Options::instance();
    mameCommand = mMameExecutable + " -listinfo > " + INFO_FILE;

    if (options->verbose)
        printf("Reading info from MAME\n");

    rc = system(mameCommand.c_str());
    if (rc < 0)
    {
        throw Exception("Could not run " + mMameExecutable);
    }

    configGenerator = new ConfigGenerator(INFO_FILE, mConfigurationName,
                                          mMameRomDirectory);

    configGenerator->generateConfig();
    delete configGenerator;
}


void
Scanner::cleanupTemporaryFiles()
{
    remove(INFO_FILE);
}
