#include <stdio.h>
extern "C" {
#include "bcd.h"
}
#include "cdbgi.h"
#include "logfile.h"

using namespace GameLauncher;
using LogFile::gl_log;


CD_Background *
CD_Background::create() throw (InitError)
{
    return new CD_Background_Impl();
}

CD_Background::~CD_Background()
{
}

CD_Background_Impl::CD_Background_Impl() throw (InitError)
{
    int rc;
    int last_track;
    int start_frame;
    int num_frames;

    rc = bcd_open();
    if (rc == 0)
    {
        gl_log->info(string("Unable to open BCD: ") + bcd_error());
        return;
    }

    // This forces BCD to re-read the disc
    (void) bcd_disc_changed();

    rc = bcd_lock(1);
    if (rc == 0)
    {
        gl_log->info(string("Unable to lock door: ") + bcd_error());
        return;
    }
    
    last_track = bcd_get_audio_info();
    if (last_track == 0)
    {
        gl_log->info("No audio tracks on CD");
        return;
    }

    rc = bcd_get_track_address(1, &start_frame, &num_frames);
    if (rc == 0)
    {
        return;
    }
    p_first_frame = start_frame;

    rc = bcd_get_track_address(last_track, &start_frame, &num_frames);
    if (rc == 0)
    {
        return;
    }
    p_last_frame = start_frame + num_frames;

    p_total_frames = p_last_frame - p_first_frame + 1;

    p_current_frame = p_first_frame;

    char buf[80];
    sprintf(buf, "Compact Disc has %d tracks", last_track);
    gl_log->info(buf);

    return;
}

CD_Background_Impl::~CD_Background_Impl()
{
    // Unlock
    (void) bcd_lock(0);
    (void) bcd_close();
    
    return;
}

void
CD_Background_Impl::poll()
{
    unsigned int        seconds;

    // If stopped, start from beginning.  Only poll once a second to
    // reduce overhead.
    seconds = p_engine->elapsed_seconds();
    if (seconds != p_elapsed_seconds)
    {
        if (bcd_audio_busy() == 0)
        {
            suspend();
            p_current_frame = p_first_frame;
            resume();
        }
        p_elapsed_seconds = seconds;
    }
}

void
CD_Background_Impl::resume()
{
    int frames_to_play;

    frames_to_play = p_last_frame - p_current_frame + 1;
    (void) bcd_play(p_current_frame, frames_to_play);

    p_elapsed_seconds = p_engine->elapsed_seconds();
    p_suspended = false;
    return;
}

void
CD_Background_Impl::suspend()
{
    p_current_frame = bcd_audio_position();
    bcd_stop();

    p_suspended = true;
    return;
}

int
CD_Background_Impl::current_frame() const
{
    if (p_suspended)
        return p_current_frame;
    else
        return bcd_audio_position();
}

int
CD_Background_Impl::total_frames() const
{
    return p_total_frames;
}

void
CD_Background_Impl::seek(int seek_frame)
{
    bool        was_running;

    was_running = false;
    if (! p_suspended)
    {
        was_running = true;
        suspend();
    }

    // Frame 0 means beginning of disc.
    if (seek_frame == 0)
    {
        p_current_frame = p_first_frame;
    }
    // This shouldn't happen, but just in case...
    else if (seek_frame > p_last_frame)
    {
        p_current_frame = p_first_frame;
    }
    else
    {
        p_current_frame = seek_frame;
    }

    if (was_running)
        resume();

    return;
}

void
CD_Background_Impl::set_volume(int volume)
{
    int bcd_volume;

    // Convert 0 - 1000 number to 0 - 255
    bcd_volume = (255 * volume) / 1000;
    (void) bcd_set_volume(bcd_volume);
}
