#include <fstream>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include "allegro.h"
#include "mp3bgi.h"
#include "fileutil.h"
#include "logfile.h"

using namespace GameLauncher;
using LogFile::gl_log;

MP3_Background *
MP3_Background::create()
    throw (InitError)
{
    return new MP3_Background_Almp3_Stream();
}

MP3_Background::~MP3_Background()
{
}

MP3_Background_Almp3_File::MP3_Background_Almp3_File()
    throw ()
    : p_playlist()
{
    p_current_mp3_index = 0;
    p_is_suspended      = true;
    p_current_mp3       = 0;
    p_mp3_data          = 0;
    p_current_position  = 0;
    p_volume            = MAX_VOLUME;
}

MP3_Background_Almp3_File::~MP3_Background_Almp3_File()
{
    close_current_mp3();
}

bool
MP3_Background_Almp3_File::open_current_mp3()
{
    PACKFILE *  file;
    string      filename;
    long        size;

    filename = get_current_mp3_name();

    size = file_size(filename);
    file = pack_fopen(filename.c_str(), F_READ);
    if (file == 0)
        return false;

    p_mp3_data = new unsigned char[size];

    pack_fread(p_mp3_data, size, file);

    pack_fclose(file);

    p_current_mp3 = almp3_create_mp3(p_mp3_data, size);
    if (p_current_mp3 == 0)
    {
        close_current_mp3();
        return false;
    }
    almp3_play_ex_mp3(p_current_mp3, BUFFER_SIZE, p_volume, PAN, SPEED,
                      NO_LOOP);
    

    return true;
}

void
MP3_Background_Almp3_File::close_current_mp3()
{
    if (p_current_mp3 != 0)
    {
        almp3_destroy_mp3(p_current_mp3);
        p_current_mp3 = 0;
    }

    if (p_mp3_data != 0)
    {
        delete p_mp3_data;
        p_mp3_data = 0;
    }
}

void
MP3_Background_Almp3_File::goto_next_mp3()
{
    p_current_mp3_index++;
    p_current_mp3_index %= p_playlist.size();
}

void
MP3_Background_Almp3_File::poll()
{
    int rc;

    if (p_is_suspended)
        return;

    rc = almp3_poll_mp3(p_current_mp3);

    if ((rc == ALMP3_POLL_PLAYJUSTFINISHED) || (rc == ALMP3_POLL_NOTPLAYING))
    {
        suspend();
        close_current_mp3();

        goto_next_mp3();

        resume();
    }
}

void
MP3_Background_Almp3_File::resume()
{
    // Stay suspended if playlist is empty.
    if (p_playlist.size() > 0)
    {
        if (p_current_mp3 == 0)
        {
            // This always opens at the beginning
            while(open_current_mp3() == false)
            {
                gl_log->warning("An error occured while opening " +
                             get_current_mp3_name());
                goto_next_mp3();
            }
        }

        if (p_current_mp3 != 0)
        {
            almp3_seek_abs_frames_mp3(p_current_mp3, p_current_position);
            p_is_suspended = false;
        }
    }
}

void
MP3_Background_Almp3_File::suspend()
{
    if (p_current_mp3 != 0)
    {
        almp3_stop_mp3(p_current_mp3);
        p_current_position = almp3_get_pos_frames_mp3(p_current_mp3);
    }
    p_is_suspended = true;
}


void
MP3_Background_Almp3_File::set_playlist(const vector<string> & playlist)
{
    p_playlist = playlist;
}

void
MP3_Background_Almp3_File::set_volume(int volume)
{
    // Convert 0 - 1000 number to 0 - 255
    p_volume = (255 * volume) / 1000;
    if (p_current_mp3 != 0)
    {
        almp3_adjust_mp3(p_current_mp3, p_volume, PAN, SPEED, NO_LOOP);
    }
    
}

int
MP3_Background_Almp3_File::get_current_position() const
{
    if (p_is_suspended)
        return p_current_position;
    else
        return almp3_get_pos_frames_mp3(p_current_mp3);
}

string
MP3_Background_Almp3_File::get_current_mp3_name() const
{
    if (p_playlist.size() == 0)
        return string("");
    else
        return p_playlist[p_current_mp3_index];
}

unsigned
MP3_Background_Almp3_File::get_current_mp3_index() const
{
    return p_current_mp3_index;
}

void
MP3_Background_Almp3_File::seek(unsigned index, int position = 0)
{
    bool        was_running;

    was_running = false;
    if (! p_is_suspended)
    {
        was_running = true;
        suspend();
    }
    close_current_mp3();

    p_current_mp3_index = 0;
    p_current_position  = 0;

    if (index < p_playlist.size())
    {
        p_current_mp3_index = index;
        p_current_position  = position;
    }

    if (was_running)
        resume();

    return;
}

void
MP3_Background_Almp3_File::skip_to_next_song()
{
    int next_mp3;
    
    next_mp3    = p_current_mp3_index + 1;
    next_mp3    %= p_playlist.size();
    seek(next_mp3);
}

void
MP3_Background_Almp3_File::start_song_over()
{
    seek(p_current_mp3_index);
}


// ----------------------------------------------------------------------


MP3_Background_Almp3_Stream::MP3_Background_Almp3_Stream()
    throw ()
    : p_playlist()
{
    p_current_mp3_index = 0;
    p_is_suspended      = true;
    p_current_mp3       = 0;
    p_mp3_file          = 0;
    p_current_position  = 0;
    p_initial_offset    = 0;
    p_volume            = MAX_VOLUME;
}

MP3_Background_Almp3_Stream::~MP3_Background_Almp3_Stream()
{
    close_current_mp3();
}

bool
MP3_Background_Almp3_Stream::open_current_mp3()
{
    string      filename;
    long        size;
    char        initial_block[BLOCK_SIZE];
    long        bytes_read;

    filename = get_current_mp3_name();

    size = file_size(filename);

    if (p_initial_offset > size)
        return false;

    p_mp3_file = pack_fopen(filename.c_str(), F_READ);
    if (p_mp3_file == 0)
        return false;

    pack_fseek(p_mp3_file, p_initial_offset);

    bytes_read = pack_fread(initial_block, BLOCK_SIZE, p_mp3_file);

    if (bytes_read < BLOCK_SIZE)
    {
        p_current_mp3 = almp3_create_mp3stream(initial_block, bytes_read,
                                               TRUE);
    }
    else
    {
        p_current_mp3 = almp3_create_mp3stream(initial_block, BLOCK_SIZE,
                                               FALSE);
    }

    if (p_current_mp3 == 0)
    {
        close_current_mp3();
        return false;
    }
    almp3_play_ex_mp3stream(p_current_mp3, BUFFER_SIZE, p_volume, PAN, SPEED);
    

    return true;
}

void
MP3_Background_Almp3_Stream::close_current_mp3()
{
    if (p_current_mp3 != 0)
    {
        almp3_destroy_mp3stream(p_current_mp3);
        p_current_mp3 = 0;
    }

    if (p_mp3_file != 0)
    {
        pack_fclose(p_mp3_file);
        p_mp3_file = 0;
    }

    p_current_position  = 0;
    p_initial_offset    = 0;
}

void
MP3_Background_Almp3_Stream::goto_next_mp3()
{
    p_current_mp3_index++;
    p_current_mp3_index %= p_playlist.size();
}

void
MP3_Background_Almp3_Stream::poll()
{
    char *  data;
    int     rc;
    long    len;

    if (p_is_suspended)
        return;

    data = (char *) almp3_get_mp3stream_buffer(p_current_mp3);

    if (data != 0)
    {
        len = pack_fread(data, BLOCK_SIZE, p_mp3_file);
        if (len < BLOCK_SIZE)
            almp3_free_mp3stream_buffer(p_current_mp3, len);
        else
            almp3_free_mp3stream_buffer(p_current_mp3, -1);
    }

    rc = almp3_poll_mp3stream(p_current_mp3);

    if ((rc == ALMP3_POLL_PLAYJUSTFINISHED) || (rc == ALMP3_POLL_NOTPLAYING))
    {
        suspend();
        close_current_mp3();

        goto_next_mp3();

        resume();
    }
}

void
MP3_Background_Almp3_Stream::resume()
{
    // Stay suspended if playlist is empty.
    if (p_playlist.size() > 0)
    {
        if (p_current_mp3 == 0)
        {
            unsigned int failed_opens;

            // Try opening the next MP3.  If we cycle through the
            // whole playlist, then just give up.
            failed_opens = 0;
            while ((failed_opens < p_playlist.size()) &&
                   (open_current_mp3() == false))
            {
                gl_log->warning("An error occured while opening " +
                             get_current_mp3_name());
                goto_next_mp3();
                failed_opens++;
            }
        }

        if (p_current_mp3 != 0)
        {
            p_is_suspended = false;
        }
    }
}

void
MP3_Background_Almp3_Stream::suspend()
{
    if (p_current_mp3 != 0)
    {
        int current_position;

        // Closing the MP3 destroys current position.  Save it, then
        // reset it after closing.  It will then be used as the
        // starting point when resumed.
        current_position = (p_initial_offset +
                            almp3_get_pos_bytes_mp3stream(p_current_mp3));
        close_current_mp3();
        p_current_position = current_position;
    }
    p_is_suspended = true;
}


void
MP3_Background_Almp3_Stream::set_playlist(const vector<string> & playlist)
{
    p_playlist = playlist;
}

void
MP3_Background_Almp3_Stream::set_volume(int volume)
{
    // Convert 0 - 1000 number to 0 - 255
    p_volume = (255 * volume) / 1000;
    if (p_current_mp3 != 0)
    {
        almp3_adjust_mp3stream(p_current_mp3, p_volume, PAN, SPEED);
    }
}

int
MP3_Background_Almp3_Stream::get_current_position() const
{
    if (p_is_suspended)
        return p_current_position;
    else
        return (p_initial_offset +
                almp3_get_pos_bytes_mp3stream(p_current_mp3));
}

string
MP3_Background_Almp3_Stream::get_current_mp3_name() const
{
    if (p_playlist.size() == 0)
        return string("");
    else
        return p_playlist[p_current_mp3_index];
}

unsigned
MP3_Background_Almp3_Stream::get_current_mp3_index() const
{
    return p_current_mp3_index;
}

void
MP3_Background_Almp3_Stream::seek(unsigned index, int position = 0)
{
    bool        was_running;

    was_running = false;
    if (! p_is_suspended)
    {
        was_running = true;
        suspend();
    }
    close_current_mp3();

    p_current_mp3_index = 0;
    p_initial_offset    = 0;

    if (index < p_playlist.size())
    {
        p_current_mp3_index = index;
        p_initial_offset    = position;
    }

    if (was_running)
        resume();

    return;
}

void
MP3_Background_Almp3_Stream::skip_to_next_song()
{
    int next_mp3;
    
    next_mp3    = p_current_mp3_index + 1;
    next_mp3    %= p_playlist.size();
    seek(next_mp3);
}

void
MP3_Background_Almp3_Stream::start_song_over()
{
    seek(p_current_mp3_index);
}
