#include <stdio.h>
#include "glaunch.h"
#include "sshot.h"
#include "menu.h"
#include "logfile.h"

using namespace GameLauncher;
using LogFile::gl_log;

int     Selection::default_unselected_body_color;
int     Selection::default_unselected_outline_color;
int     Selection::default_selected_body_color;
int     Selection::default_selected_outline_color;

Selection::Selection(Engine * engine, FONT * new_font, BITMAP * icon)
    : p_selected(false), p_state(STOPPED)
{
    bool        try_vram;

    p_engine    = engine;

    // Use global font
    p_font      = new_font;
    p_pixel_height      = text_height(p_font) + OUTLINE_PADDING;
    
    p_unselected_body_color     = default_unselected_body_color;
    p_unselected_outline_color  = default_unselected_outline_color;
    p_selected_body_color       = default_selected_body_color;
    p_selected_outline_color    = default_selected_outline_color;

    p_use_rle_sprite    = false;
    p_rle_sprite        = 0;
    p_bitmap    = 0;
    p_x         = 0;
    p_y         = 0;
    p_x_velocity        = 0;
    p_y_velocity        = 0;
    p_clip_top          = 0;
    p_clip_bottom       = SCREEN_H;

#if USE_VRAM
    try_vram = ((p_engine->animation_method() == PAGE_FLIPPING)
                && ((gfx_capabilities & GFX_HW_VRAM_BLIT_MASKED) != 0));
#else
    try_vram = false;
#endif

    if (try_vram)
    {
        p_bitmap = create_video_bitmap(SCREEN_W/2 - 1, p_pixel_height);
        if (p_bitmap != 0)
            gl_log->info("VRAM bitmap for selection");
    }

    if (p_bitmap == 0)
    {
        if (try_vram)
            gl_log->info("No VRAM: memory bitmap for selection");
        else
            gl_log->info("No HW blit: memory bitmap for selection");

        p_bitmap = create_bitmap(SCREEN_W/2 - 1, p_pixel_height);
        p_use_rle_sprite = true;
    }

    p_icon  = icon;
    set_text("", false);
    update_bitmap();
}

Selection::~Selection()
{
    if (p_rle_sprite != 0)
        destroy_rle_sprite(p_rle_sprite);
    if (p_bitmap != 0)
        destroy_bitmap(p_bitmap);

    return;
}
void
Selection::set_text(const string & text, bool show_icon = false)
{
    char *      end;
    int         width;

    strncpy(p_text, text.c_str(), MAX_TEXT_LEN);
    p_pixel_width = text_length(p_font, p_text) + OUTLINE_PADDING;
    
    width = p_bitmap->w;
    // Save space for screen shot icon
    if (show_icon && (p_icon != 0))
    {
        p_show_icon = true;
        width -= p_icon->w;
    }
    else
    {
        p_show_icon = false;
    }

    // Chop off a character at a time to make sure it fits
    end = &p_text[strlen(p_text) - 1];
    while (p_pixel_width > width)
    {
        *end = '\0';
        end--;
        p_pixel_width = text_length(p_font, p_text) + OUTLINE_PADDING;
    }

    if (p_show_icon)
        p_pixel_width += p_icon->w;

    update_bitmap();

    return;
}

void
Selection::update_bitmap()
{
    int bg_color;
    int body_color;
    int outline_color;
    
    if (p_bitmap == 0)
        return;

    bg_color = makecol(255, 0, 255);
    clear_to_color(p_bitmap, bg_color);
    
    if (p_selected)
    {
        body_color      = p_selected_body_color;
        outline_color   = p_selected_outline_color;
    }
    else
    {
        body_color      = p_unselected_body_color;
        outline_color   = p_unselected_outline_color;
    }
    outline_textout(p_bitmap, p_font, p_text,
                    0, 0, body_color, outline_color);

    if (p_show_icon)
    {
        // Put screen shot icon on the right
        draw_sprite(p_bitmap, p_icon, (p_pixel_width - p_icon->w),
                    (p_icon->h / 2));
    }

    if (p_use_rle_sprite)
    {
        if (p_rle_sprite != 0)
        {
            destroy_rle_sprite(p_rle_sprite);
            p_rle_sprite = 0;
        }

        p_rle_sprite = get_rle_sprite(p_bitmap);
    }
}

void
Selection::select()
{
    if (p_selected != true)
    {
        p_selected = true;
        update_bitmap();
    }
}

void
Selection::unselect()
{
    if (p_selected != false)
    {
        p_selected = false;
        update_bitmap();
    }
}

void
Selection::scroll_to(int x, int y, int milliseconds)
{
    fix       num_game_ticks;

    num_game_ticks = (milliseconds * p_engine->updates_per_second()) / 1000;

    p_scroll_dest_x = x;
    p_scroll_dest_y = y;
    p_x_velocity = (p_scroll_dest_x - p_x) / num_game_ticks;
    p_y_velocity = (p_scroll_dest_y - p_y) / num_game_ticks;

    // If destination is so close such that the velocity is zero (with
    // rounding), stay stopped
    if ((p_x_velocity == 0) && (p_y_velocity == 0))
        p_state = STOPPED;
    else
        p_state = SCROLLING;
}

void
Selection::move_to(int x, int y)
{
    p_x = x;
    p_y = y;
    p_state = STOPPED;
}

int
Selection::x()
{
    return p_x;
}

int
Selection::y()
{
    return p_y;
}

int
Selection::s_x()
{
    return p_scroll_dest_x;
}

int
Selection::s_y()
{
    return p_scroll_dest_y;
}

bool
Selection::is_stopped()
{
    return (p_state == STOPPED);
}

void
Selection::set_clip(int top, int bottom)
{
    p_clip_top          = top;
    p_clip_bottom       = bottom;
    return;
}

void
Selection::draw(BITMAP * frame)
{
    fix top;
    fix bottom;
    fix height;
    fix skip;

    top         = p_y;
    bottom      = top + p_pixel_height - 1;

    if (top < p_clip_top)
        top = p_clip_top;
    if (bottom > p_clip_bottom)
        bottom = p_clip_bottom;

    skip = top - p_y;
    height = bottom - top + 1;

#if 0
    if ((p_use_rle_sprite) && (skip == 0) && (height == p_pixel_height))
    {
        draw_rle_sprite(frame, p_rle_sprite, (p_x - p_pixel_width/2), top);
    }
    else
#endif
    {
        masked_blit(p_bitmap, frame, 0, skip, (p_x - p_pixel_width/2), top,
                    p_pixel_width, height);
    }

    return;
}

bool
Selection::update()
{
    if (p_state == STOPPED)
        return false;
    else if (p_state == SCROLLING)
    {
        bool    moving_right;
        bool    moving_down;

        // Will we reach or pass our final destination?
        moving_right    = (p_x_velocity > 0);
        moving_down     = (p_y_velocity > 0);

        if ((moving_right && ((p_x + p_x_velocity) >= p_scroll_dest_x)) ||
            ((! moving_right) && ((p_x + p_x_velocity) <= p_scroll_dest_x)))
        {
            p_x = p_scroll_dest_x;
        }
        else
            p_x += p_x_velocity;

        if ((moving_down && ((p_y + p_y_velocity) >= p_scroll_dest_y)) ||
            ((! moving_down) && (p_y + p_y_velocity) <= p_scroll_dest_y))
        {
            p_y = p_scroll_dest_y;
        }
        else
            p_y += p_y_velocity;

        if ((p_x == p_scroll_dest_x) && (p_y == p_scroll_dest_y))
           p_state = STOPPED;
    }
    
    return true;
}

Menu::Menu(Engine * engine, Screen_Shot * screen_shot, FONT * new_font,
           BITMAP * screen_shot_icon)
    : p_selections(0), p_all_selections(),
      p_screen_shot(screen_shot),
      p_state(INIT)
{
    p_engine            = engine;
    p_font              = new_font;
    p_font_height       = text_height(p_font) + Selection::OUTLINE_PADDING;

    // Default samples and icons to nothing;
    p_next_emu          = 0;
    p_previous_emu      = 0;
    p_selection_moved   = 0;
    p_game_chosen       = 0;

    p_screen_shot_icon  = screen_shot_icon;

    // Setup selection style used with all selections
    Selection::default_unselected_body_color    = makecol(192, 192, 192);
    Selection::default_unselected_outline_color = makecol(  0,   0,   0);
    Selection::default_selected_body_color      = makecol(  0,   0,   0);
    Selection::default_selected_outline_color   = makecol(192, 192,  64); 

    p_selected_index = 0;
    p_first_shown_index = 0;

    // Leave room at the top for title, sub title, and separator
    p_y_top = p_font_height * 2 + p_separator_height;

    p_num_shown = (SCREEN_H - p_y_top) / p_font_height;
    p_y_bottom = p_y_top + p_font_height * p_num_shown;

    unsigned int        i;
    int x, y;
    int clip_top, clip_bottom;
    GEContainer *       root;

    root        = p_engine->get_root_container();
    clip_top    = p_y_top;
    clip_bottom = p_y_bottom;
    y           = p_y_top;
    for (i = 0; i < p_num_shown; i++)
    {
        Selection *     selection;

        selection = new Selection(p_engine, p_font, p_screen_shot_icon);
        if ((i % 2) == 0)
            x = SCREEN_W * 3/2;
        else
            x = 0 - SCREEN_W/2;
        
        // Setup selection parameters before enabling it.
        selection->move_to(x, y);
        selection->set_clip(clip_top, clip_bottom);

        root->add_updatable(selection, 500);
        root->add_drawable(selection, 500);

        p_selections.push_back(selection);
        y += p_font_height;
    }

    // Create the reserves
    for (i = 0; i < p_num_shown; i++)
    {
        Selection *     selection;

        selection = new Selection(p_engine, p_font, p_screen_shot_icon);

        selection->set_clip(clip_top, clip_bottom);
        selection->move_to(SCREEN_W * 3/2, 0);         // Move off screen

        root->add_updatable(selection, 500);
        root->add_drawable(selection, 500);

        p_scroll_reserves.push_back(selection);
    }

    p_selected = p_selections.begin();
    (*p_selected)->select();

    // Create the "selections" for the title and sub-title
    p_title = new Selection(p_engine, p_font, 0);
    p_title->move_to(SCREEN_W/2, 0);
    root->add_updatable(p_title, 500);
    root->add_drawable(p_title, 500);

    p_sub_title = new Selection(p_engine, p_font, 0);
    p_sub_title->move_to(SCREEN_W/2, p_font_height);
    root->add_updatable(p_sub_title, 500);
    root->add_drawable(p_sub_title, 500);

    p_menu_callback     = 0;

    p_screen_shot_brightness    = 40;
    p_screen_shot_delay         = 0;

    p_init_timer        = 0;
    p_done_timer        = 0;
    p_screen_shot_timer = p_engine->get_timer_factory()->create_timer();

    p_key_initial_delay = 250;
    p_key_repeat_delay  = 100;

    p_up_key            = 0;
    p_down_key          = 0;
    p_page_up_key       = 0;
    p_page_down_key     = 0;
    p_play_game_key     = 0;
    p_next_emu_key      = 0;
    p_previous_emu_key  = 0;
    p_quit_key          = 0;
    p_screen_shot_key   = 0;

    install_key_buffer();

    return;
}

Menu::~Menu()
{
    Selection_List::iterator i;

    for (i = p_selections.begin(); i != p_selections.end(); i++)
    {
        delete *i;
    }
    for (i = p_scroll_reserves.begin(); i != p_scroll_reserves.end(); i++)
    {
        delete *i;
    }

    if (p_init_timer != 0)
        delete p_init_timer;

    if (p_done_timer != 0)
        delete p_done_timer;

    delete p_screen_shot_timer;

    return;
}

static const int KEY_BUFFER_SIZE        = 64;

struct Key_Buffer
{
    volatile int        start;
    volatile int        end;
    volatile int        size;
    volatile int        buffer[KEY_BUFFER_SIZE];
};

static volatile Key_Buffer      key_buffer;

static void
Menu_kb_handler(int scancode)
{
    
    if (key_buffer.size == KEY_BUFFER_SIZE)
        // Buffer is full
        return;
    
    key_buffer.buffer[key_buffer.start] = scancode;
    key_buffer.size++;

    key_buffer.start++;
    key_buffer.start  %= KEY_BUFFER_SIZE;

    return;
}
END_OF_FUNCTION(Menu_kb_handler);

bool
Menu::key_buffer_empty()
{
    return (key_buffer.size == 0);
}

int
Menu::next_key_in_buffer()
{
    int key;

    asm volatile ("CLI");
    if (key_buffer.size == 0)
        key = 0;
    else
    {
        key_buffer.end++;
        key_buffer.end %= KEY_BUFFER_SIZE;
        key = key_buffer.buffer[key_buffer.end];
        key_buffer.size--;
    }
    asm volatile ("STI");

    return key;
}

void
Menu::clear_key_buffer()
{
    key_buffer.size     = 0;
    key_buffer.start    = 0;
    key_buffer.end      = KEY_BUFFER_SIZE - 1;

    return;
}

void
Menu::install_key_buffer()
{
    LOCK_VARIABLE(key_buffer);
    LOCK_FUNCTION(Menu_kb_handler);

    clear_key_buffer();

    keyboard_lowlevel_callback = Menu_kb_handler;

    return;
}

void
Menu::set_selection(int num, char * selection)
{
    sprintf(selection, "Blah selection %0*d", num, num);
    return;
}

bool
Menu::are_all_selections_stopped()
{
    Selection_List::iterator   i;
    bool    all_stopped;

    all_stopped = true;
    for (i = p_selections.begin(); i != p_selections.end(); i++)
    {
        all_stopped = all_stopped && (*i)->is_stopped();
    }
    return all_stopped;
}

void
Menu::move_selection_up(unsigned int shift, unsigned int scroll)
{
    unsigned int i;
    int         scroll_ms;
    Selection_List      moved_selections;

    scroll_ms = p_vert_scroll_ms * ((scroll + 1)/ 2);

    // First, perform the scroll
    for (i = 0; i < scroll; i++)
    {
        Selection *     selection;
        unsigned int    next_idx;

        // Remove last item, but stick it on the reserves list
        selection = p_selections.back();
        p_selections.pop_back();

        p_scroll_reserves.push_front(selection);
        moved_selections.push_front(selection);

        // Pull a selection of the reserves list, setup the new text,
        // and move it to the bottom.
        selection = p_scroll_reserves.back();
        p_scroll_reserves.pop_back();

        next_idx = p_selected_index - i - 1;

        if (p_menu_callback->screen_shot_available(next_idx))
            selection->set_text(p_all_selections[next_idx], true);
        else
            selection->set_text(p_all_selections[next_idx], false);
        
        selection->move_to(SCREEN_W/2, p_y_top - ((i+1)*p_font_height));
        p_selections.push_front(selection);

        p_first_shown_index--;
        p_engine->poll_background_task();
    }

    // Next, perform shift.  Must be done in a loop becuase
    // p_selections is not random access.
    for (i = 0; i < shift; i++)
    {
        p_selected_index--;
        p_selected--;
    }

    // Move positions
    if (scroll > 0)
    {
        Selection_List::iterator    j;
        int x;
        int y;

        y = p_y_top;
        for (j = p_selections.begin(); j != p_selections.end(); j++)
        {
            x = (*j)->x();
            if (scroll > 0)
                (*j)->scroll_to(x, y, scroll_ms);
            y += p_font_height;
        }

        // The first `scroll' items on the reserve list need to be
        // scrolled off the screen
        for (i = 0, j = p_scroll_reserves.begin();
             i < scroll;
             i++, j++)
        {
            x = (*j)->x();
            (*j)->scroll_to(x, y, scroll_ms);
            y += p_font_height;
        }

        p_state = SCROLLING_UP_OR_DOWN;
    }

    return;
}

void
Menu::move_selection_down(unsigned int shift, unsigned int scroll)
{
    unsigned int i;
    int         scroll_ms;

    scroll_ms = p_vert_scroll_ms * ((scroll + 1)/ 2);

    // First, perform the scroll
    for (i = 0; i < scroll; i++)
    {
        Selection *     selection;
        unsigned int    next_idx;

        // Remove first item, but stick it on the back of the reserves
        // list
        selection = p_selections.front();
        p_selections.pop_front();
        p_scroll_reserves.push_back(selection);

        // Pull a selection off the front of the reserves list, setup
        // the new text, and move it to the bottom.
        selection = p_scroll_reserves.front();
        p_scroll_reserves.pop_front();

        next_idx = p_selected_index + i + 1;
        if (p_menu_callback->screen_shot_available(next_idx))
            selection->set_text(p_all_selections[next_idx], true);
        else
            selection->set_text(p_all_selections[next_idx], false);
        selection->move_to(SCREEN_W/2,
                           p_y_bottom + (i*p_font_height));
        p_selections.push_back(selection);

        p_first_shown_index++;
        p_engine->poll_background_task();
    }

    // Next, perform shift.  Must be done in a loop becuase
    // p_selections is not random access.
    for (i = 0; i < shift; i++)
    {
        p_selected_index++;
        p_selected++;
    }

    // Move positions
    if (scroll > 0)
    {
        Selection_List::reverse_iterator    j;
        int x;
        int y;

        y = p_y_bottom - p_font_height;
        for (j = p_selections.rbegin(); j != p_selections.rend(); j++)
        {
            x = (*j)->x();
            if (scroll > 0)
                (*j)->scroll_to(x, y, scroll_ms);
            y -= p_font_height;
        }

        // The last `scroll' items on the reserve list need to be
        // scrolled off the screen
        for (i = 0, j = p_scroll_reserves.rbegin();
             i < scroll;
             i++, j++)
        {
            x = (*j)->x();
            (*j)->scroll_to(x, y, scroll_ms);
            y -= p_font_height;
        }

        p_state = SCROLLING_UP_OR_DOWN;
    }

    return;
}

void
Menu::draw(BITMAP * frame)
{
    int white;
    int black;
    int x_midpoint;
    int y_midpoint;
    int half_width;
    int half_height;
    int outline;
    int x1, y1, x2, y2;

    if (p_state == DISPLAYING_SCREEN_SHOT)
    {
        return;
    }

    black = makecol(  0,   0,   0);
    white = makecol(255, 255, 255);

    x_midpoint  = SCREEN_W/2;
    y_midpoint  = p_font_height*2 + p_separator_height/2;
    outline     = 2;

    half_width = SCREEN_W/4;
    half_height = 1;

    x1 = x_midpoint - half_width  - outline;
    y1 = y_midpoint - half_height - outline;
    x2 = x_midpoint + half_width  + outline;
    y2 = y_midpoint + half_height + outline;

    rectfill(frame, x1, y1, x2, y2, black);

    x1 += outline;
    y1 += outline;
    x2 -= outline;
    y2 -= outline;

    rectfill(frame, x1, y1, x2, y2, white);

    return;
}

bool
Menu::update()
{
    bool        updated;

    updated = false;

    if (p_state == TAKING_ORDERS)
    {
        return take_orders();
    }

    if (p_state == INIT)
    {
        p_init_timer = p_engine->get_timer_factory()->create_timer();
        p_init_timer->set(1000);
        p_state = INIT_WAIT;
    }

    if (p_state == INIT_WAIT)
    {
        if (p_init_timer->expired())
        {
            Selection_List::iterator    i;
            int scroll_x;
            int scroll_y;

            scroll_x = SCREEN_W / 2;
            scroll_y = p_y_top;
            for (i = p_selections.begin(); i != p_selections.end(); i++)
            {
                (*i)->scroll_to(scroll_x, scroll_y, p_lr_scroll_ms);
                scroll_y += p_font_height;
            }
            
            delete p_init_timer;
            p_init_timer = 0;

            p_state = SCROLLING_ON;
            updated = true;
        }
    }

    if (p_state == SCROLLING_ON)
    {
        if (are_all_selections_stopped())
        {
            p_menu_callback->update_screen_shot(p_selected_index,
                                                p_screen_shot_brightness);

            p_state = TAKING_ORDERS;
            updated = true;
        }
    }

    if (p_state == SCROLLING_UP_OR_DOWN)
    {
        if (are_all_selections_stopped())
        {
            p_screen_shot_timer->set(p_screen_shot_delay);

            p_state = TAKING_ORDERS;
            updated = take_orders();
        }
    }

    if ((p_state == SCROLLING_OFF_RIGHT) || (p_state == SCROLLING_OFF_LEFT))
    {
        if (are_all_selections_stopped())
        {
            Selection_List::iterator i;
            int scroll_x;
            int scroll_y;

            if (p_state == SCROLLING_OFF_LEFT)
            {
                p_menu_callback->previous_emu(p_first_shown_index,
                                              p_selected_index);
            }
            else
            {
                p_menu_callback->next_emu(p_first_shown_index,
                                          p_selected_index);
            }

            scroll_x = SCREEN_W / 2;
            for (i = p_selections.begin(); i != p_selections.end(); i++)
            {
                scroll_y = (*i)->y();
                if ((*i)->x() < 0)
                    (*i)->move_to(SCREEN_W * 3/2, scroll_y);
                else
                    (*i)->move_to(0 - SCREEN_W/2, scroll_y);
                    
                (*i)->scroll_to(scroll_x, scroll_y, p_lr_scroll_ms);
            }
            p_title->scroll_to(scroll_x, 0, p_lr_scroll_ms);
            p_sub_title->scroll_to(scroll_x, p_font_height, p_lr_scroll_ms);

            p_state = SCROLLING_ON;
            updated = true;
        }
    }

    if (p_state == DISPLAYING_SCREEN_SHOT)
    {
        if (p_screen_shot_key->is_pressed())
        {
            Selection_List::iterator i;
            GEContainer *       root;

            root = p_engine->get_root_container();
            for (i = p_selections.begin(); i != p_selections.end(); i++)
            {
                root->add_drawable(*i, 500);
            }

            root->add_drawable(p_title, 500);
            root->add_drawable(p_sub_title, 500);

            p_menu_callback->update_screen_shot(p_selected_index,
                                                p_screen_shot_brightness);

            p_state = TAKING_ORDERS;
            updated = true;
        }
    }

    if (p_state == DONE_WAIT)
    {
        if (are_all_selections_stopped())
        {
            p_screen_shot->set_to_black();
            p_done_timer = p_engine->get_timer_factory()->create_timer();
            p_done_timer->set(p_done_delay);

            p_state = DONE;
            updated = true;
        }
    }
    
    if (p_state == DONE)
    {
        if (p_done_timer->expired())
        {
            delete p_done_timer;
            p_done_timer = 0;

            p_engine->stop();

            p_state = STOPPED;
            updated = true;
        }
    }

    if (p_state == STOPPED)
    {
        // Do nothing.... we're stopped. :-)
    }

    return updated;
}

bool
Menu::take_orders()
{
    bool        updated;

    updated = false;

    if (p_next_emu_key->is_pressed())
    {
        Selection_List::iterator i;
        int     scroll_x;
        int     scroll_y;

        scroll_y = p_y_top;
        for (i = p_selections.begin(); i != p_selections.end(); i++)
        {
            scroll_x = SCREEN_W * 3/2;
            (*i)->scroll_to(scroll_x, scroll_y, p_lr_scroll_ms);
            scroll_y += p_font_height;
        }
        p_title->scroll_to(SCREEN_W/2, 0 - p_font_height*2, p_lr_scroll_ms);
        p_sub_title->scroll_to(SCREEN_W/2, 0 - p_font_height,
                               p_lr_scroll_ms);

        p_state = SCROLLING_OFF_RIGHT;
        play_sample(p_next_emu, 255, 127, 1000, FALSE);
        updated = true;
    }

    if (p_previous_emu_key->is_pressed())
    {
        Selection_List::iterator i;
        int     scroll_x;
        int     scroll_y;

        scroll_y = p_y_top;
        for (i = p_selections.begin(); i != p_selections.end(); i++)
        {
            scroll_x = 0 - SCREEN_W/2;
            (*i)->scroll_to(scroll_x, scroll_y, p_lr_scroll_ms);
            scroll_y += p_font_height;
        }
        p_title->scroll_to(SCREEN_W/2, 0 - p_font_height*2, p_lr_scroll_ms);
        p_sub_title->scroll_to(SCREEN_W/2, 0 - p_font_height,
                               p_lr_scroll_ms);

        p_state = SCROLLING_OFF_LEFT;
        play_sample(p_previous_emu, 255, 127, 1000, FALSE);
        updated = true;
    }

    if (p_play_game_key->is_pressed())
    {
        Selection_List::iterator i;
        int     scroll_x;
        int     scroll_y;
        int     j;

        scroll_y = p_y_top;
        for (j = 0, i = p_selections.begin();
             i != p_selections.end();
             j++, i++)
        {
            // Center current selection
            if (i == p_selected)
            {
                (*i)->scroll_to(SCREEN_W/2, SCREEN_H/2, p_lr_scroll_ms);
            }
            else
            {
                if ((j % 2) == 0)
                    scroll_x = SCREEN_W * 3 / 2;
                else
                    scroll_x = 0 - SCREEN_W/2;
                (*i)->scroll_to(scroll_x, scroll_y, p_lr_scroll_ms);
            }
            scroll_y += p_font_height;
        }
        p_done_delay = 1000;
        play_sample(p_game_chosen, 255, 127, 1000, FALSE);

        p_menu_callback->select(p_first_shown_index, p_selected_index);

        p_state = DONE_WAIT;
        updated = true;
    }
    
    if (p_quit_key->is_pressed())
    {
        Selection_List::iterator i;
        int     scroll_x;
        int     scroll_y;
        int     j;

        scroll_y = p_y_top;
        for (j = 0, i = p_selections.begin();
             i != p_selections.end();
             j++, i++)
        {
            if ((j % 2) == 0)
                scroll_x = SCREEN_W * 3 / 2;
            else
                scroll_x = 0 - SCREEN_W/2;
            (*i)->scroll_to(scroll_x, scroll_y, p_lr_scroll_ms);
            scroll_y += p_font_height;
        }
        p_done_delay = 0;
        play_sample(p_quit, 255, 127, 1000, FALSE);

        p_menu_callback->quit(p_first_shown_index, p_selected_index);

        p_state = DONE_WAIT;
        updated = true;
    }

    if (p_screen_shot_key->is_pressed() &&
        (p_menu_callback->screen_shot_available(p_selected_index)))
    {
        Selection_List::iterator i;
        GEContainer *   root;

        root = p_engine->get_root_container();
        for (i = p_selections.begin(); i != p_selections.end(); i++)
        {
            root->remove_drawable(*i);
        }
        root->remove_drawable(p_title);
        root->remove_drawable(p_sub_title);

        p_menu_callback->update_screen_shot(p_selected_index, 100);

        clear_key_buffer();

        p_state = DISPLAYING_SCREEN_SHOT;
        updated = true;
    }

    unsigned int    move_up;
    unsigned int    move_down;
    unsigned int    scroll_up;
    unsigned int    scroll_down;
    unsigned int    last_shown_index;

    move_up         = 0;
    move_down       = 0;
    scroll_up       = 0;
    scroll_down     = 0;

    // Calculate last_shown_index, clipping if necessary
    last_shown_index = p_first_shown_index + p_selections.size() - 1;
    if (last_shown_index >= p_all_selections.size())
        last_shown_index = p_all_selections.size() - 1;
            

    if (p_down_key->is_pressed())
    {
        move_down = 1;
        if (p_selected_index == last_shown_index)
            scroll_down = 1;
    }

    if (p_up_key->is_pressed())
    {
        move_up = 1;
        if (p_selected_index == p_first_shown_index)
            scroll_up = 1;
    }

    if (p_page_up_key->is_pressed())
    {
        if (p_selected_index == p_first_shown_index)
        {
            // At top, scroll up a page.
            move_up = p_selections.size() - 1;
            scroll_up = p_selections.size() - 1;
        }
        else
        {
            // Move to first shown.
            move_up = p_selected_index - p_first_shown_index;
        }
    }

    if (p_page_down_key->is_pressed())
    {
        if (p_selected_index == last_shown_index)
        {
            // At end, scroll down a page.
            move_down = p_selections.size() - 1;
            scroll_down = p_selections.size() - 1;
        }
        else
        {
            // Move to last shown
            move_down = last_shown_index - p_selected_index;
        }
    }

    // Any wierd combination where up and down movement is
    // involved results in no effect
    if ((move_up > 0) && (move_down > 0))
    {
        move_down       = 0;
        move_up         = 0;
    }

    // Make sure we don't move up past the beginning.
    if ((move_up > 0) && (p_selected_index < move_up))
    {
        move_up     = p_selected_index;
        scroll_up   = p_selected_index;
    }

    // Make sure we don't move down past the end.
    if ((move_down > 0) &&
        ((p_selected_index + move_down) > (p_all_selections.size() - 1)))
    {
        move_down   = p_all_selections.size() - 1 - p_selected_index;
        scroll_down = p_all_selections.size() - 1 - p_selected_index;
    }


    if ((move_up > 0) || (move_down > 0))
    {
        gl_log->debug("Move");
        (*p_selected)->unselect();
        if (move_up > 0)
        {
            move_selection_up(move_up, scroll_up);
        }
        else if (move_down > 0)
        {
            move_selection_down(move_down, scroll_down);
        }
        (*p_selected)->select();

        p_screen_shot_timer->set(p_screen_shot_delay);
        play_sample(p_selection_moved, 255, 127, 1000, FALSE);
        updated = true;
    }
    
    if (p_screen_shot_timer->expired())
    {
        p_menu_callback->update_screen_shot(p_selected_index,
                                            p_screen_shot_brightness);
        updated = true;
    }

    return updated;
}

void
Menu::set_menu(const string & title, const string & sub_title,
               const vector<string> & all_selections,
               Menu_Callback * callback, unsigned int first_shown_index,
               unsigned int selected_index)
{
    p_all_selections    = all_selections;
    p_first_shown_index = first_shown_index;
    p_selected_index    = selected_index;
    p_menu_callback     = callback;

    Selection_List::iterator    i;
    unsigned j;
    int y;


    p_title->set_text(title + ":");
    p_sub_title->set_text(sub_title);

    (*p_selected)->unselect();
    j = p_first_shown_index;
    y = p_y_top;
    for (i = p_selections.begin(); i != p_selections.end(); i++)
    {
        if (j < p_all_selections.size())
        {
            if (p_menu_callback->screen_shot_available(j))
                (*i)->set_text(p_all_selections[j], true);
            else
                (*i)->set_text(p_all_selections[j], false);

            if (j == p_selected_index)
            {
                p_selected = i;
                (*p_selected)->select();
            }
        }
        else
        {
            (*i)->set_text("", false);
        }
        
        y += p_font_height;
        j++;
    }
}

void
Menu::set_key_repeat_delay(int initial_delay, int repeat_delay)
{
    p_key_initial_delay = initial_delay;
    p_key_repeat_delay  = repeat_delay;
}

void
Menu::free_key(Keypress_Watcher * watcher)
{
    if (watcher != 0)
    {
        delete watcher;
        watcher = 0;
    }
}

Keypress_Watcher *
Menu::get_repeat_key(int scancode)
{
    Keypress_Watcher * watcher;
    watcher = 
        p_engine->get_keypress_factory()->create_keypress_watcher(scancode);
    watcher->set_repeat(p_key_initial_delay, p_key_repeat_delay);
    return watcher;
}

Keypress_Watcher *
Menu::get_one_time_key(int scancode)
{
    Keypress_Watcher * watcher;
    watcher = 
        p_engine->get_keypress_factory()->create_keypress_watcher(scancode);
    watcher->set_one_time();
    return watcher;
}

void
Menu::set_up_key(int scancode)
{
    free_key(p_up_key);
    p_up_key = get_repeat_key(scancode);
}

void
Menu::set_down_key(int scancode)
{
    free_key(p_down_key);
    p_down_key = get_repeat_key(scancode);
}

void
Menu::set_page_up_key(int scancode)
{
    free_key(p_page_up_key);
    p_page_up_key = get_repeat_key(scancode);
}

void
Menu::set_page_down_key(int scancode)
{
    free_key(p_page_down_key);
    p_page_down_key = get_repeat_key(scancode);
}

void
Menu::set_play_game_key(int scancode)
{
    free_key(p_play_game_key);
    p_play_game_key = get_one_time_key(scancode);
}

void
Menu::set_next_emu_key(int scancode)
{
    free_key(p_next_emu_key);
    p_next_emu_key = get_one_time_key(scancode);
}

void
Menu::set_previous_emu_key(int scancode)
{
    free_key(p_previous_emu_key);
    p_previous_emu_key = get_one_time_key(scancode);
}

void
Menu::set_quit_key(int scancode)
{
    free_key(p_quit_key);
    p_quit_key = get_one_time_key(scancode);
}

void
Menu::set_screen_shot_key(int scancode)
{
    free_key(p_screen_shot_key);
    p_screen_shot_key = get_one_time_key(scancode);
}

SAMPLE *
Menu::next_emu_sample()
{
    return p_next_emu;
}

void
Menu::next_emu_sample(SAMPLE * new_sample)
{
    p_next_emu = new_sample;
}

SAMPLE *
Menu::previous_emu_sample()
{
    return p_previous_emu;
}

void
Menu::previous_emu_sample(SAMPLE * new_sample)
{
    p_previous_emu = new_sample;
}

SAMPLE *
Menu::selection_moved_sample()
{
    return p_selection_moved;
}

void
Menu::selection_moved_sample(SAMPLE * new_sample)
{
    p_selection_moved = new_sample;
}

SAMPLE *
Menu::game_chosen_sample()
{
    return p_game_chosen;
}

void
Menu::game_chosen_sample(SAMPLE * new_sample)
{
    p_game_chosen = new_sample;
}


SAMPLE *
Menu::quit_sample()
{
    return p_quit;
}

void
Menu::screen_shot_delay(int delay)
{
    p_screen_shot_delay = delay;
}

void
Menu::screen_shot_brightness(int brightness_percent)
{
    p_screen_shot_brightness = brightness_percent;
}

void
Menu::quit_sample(SAMPLE * new_sample)
{
    p_quit = new_sample;
}

void outline_textout(BITMAP * bmp, FONT * f, char * s,
                     int x, int y, int main_color, int outline_color)
{
    int offsets[][2] = 
    {
        { 2,  1},
        { 2,  0},
        { 2, -1},
        {-2,  1},
        {-2,  0},
        {-2, -1},
        { 1,  2},
        { 0,  2},
        {-1,  2},
        { 1, -2},
        { 0, -2},
        {-1, -2},
        { 0,  0}
    };
    int i, t_height, t_width, trans_col;
    BITMAP *    current_text;

    t_height    = text_height(f) + 4;
    t_width     = text_length(f, s) + 4;
    text_mode(-1);

    trans_col       = makecol(255, 0, 255);
    current_text    = create_bitmap(t_width, t_height);
    clear_to_color(current_text, trans_col);
    textout_centre(current_text, f, s, t_width/2, 2, outline_color);

    i = 0;
    while (! ((offsets[i][0] == 0) && (offsets[i][1] == 0)))
    {
        draw_sprite(bmp, current_text, x + offsets[i][0],
                    y + offsets[i][1]);
        i++;
    }

    clear_to_color(current_text, trans_col);
    textout_centre(current_text, f, s, t_width/2, 2, main_color);
    draw_sprite(bmp, current_text, x, y);

    destroy_bitmap(current_text);
    return;
}


#if 0
Menu_Generator::Menu_Generator(Engine * engine, Menu * menu)
    : p_engine(engine), p_menu(menu), p_menus(4)
{
    int         i;
    char        selection[200];

    p_menus[0].title            = "MAME";
    p_menus[0].sub_title        = "Arcade Machines";
    p_menus[0].first_shown_index = 0;
    p_menus[0].selected_index   = 0;
    p_menus[0].games.reserve(120);
    for (i = 0; i < 120; i++)
    {
        set_text(selection, p_menus[0].title.c_str(), i);
        p_menus[0].games.push_back(string(selection));
    }

    p_menus[1].title            = "Nesticle";
    p_menus[1].sub_title        = "Nintendo";
    p_menus[1].first_shown_index = 0;
    p_menus[1].selected_index   = 0;
    p_menus[1].games.reserve(55);
    for (i = 0; i < 55; i++)
    {
        set_text(selection, p_menus[1].title.c_str(), i);
        p_menus[1].games.push_back(string(selection));
    }

    p_menus[2].title            = "Stella";
    p_menus[2].sub_title        = "Atari 2600";
    p_menus[2].first_shown_index = 0;
    p_menus[2].selected_index   = 0;
    p_menus[2].games.reserve(300);
    for (i = 0; i < 300; i++)
    {
        set_text(selection, p_menus[2].title.c_str(), i);
        p_menus[2].games.push_back(string(selection));
    }

    p_menus[3].title            = "Genecyst";
    p_menus[3].sub_title        = "Sega Genesis";
    p_menus[3].first_shown_index = 0;
    p_menus[3].selected_index   = 0;
    p_menus[3].games.reserve(12);
    for (i = 0; i < 12; i++)
    {
        set_text(selection, p_menus[3].title.c_str(), i);
        p_menus[3].games.push_back(string(selection));
    }
    
    p_current_menu = 0;
    setup_menu();
    return;
}

Menu_Generator::~Menu_Generator()
{
}

void
Menu_Generator::next_emu(unsigned int first_shown_index,
                         unsigned int selected_index)
{
    p_menus[p_current_menu].first_shown_index = first_shown_index;
    p_menus[p_current_menu].selected_index = selected_index;
    p_current_menu++;
    p_current_menu %= p_menus.size();
    setup_menu();
    return;
}

void
Menu_Generator::previous_emu(unsigned int first_shown_index,
                             unsigned int selected_index)
{
    p_menus[p_current_menu].first_shown_index = first_shown_index;
    p_menus[p_current_menu].selected_index = selected_index;
    if (p_current_menu == 0)
        p_current_menu = p_menus.size() - 1;
    else
        p_current_menu--;
    setup_menu();
    return;
}

void
Menu_Generator::select(unsigned int first_shown_index,
                       unsigned int selected_index)
{
    cout << "Selected " << selected_index << " from "
         << p_menus[p_current_menu].title << " menu\n";
    return;
}

void
Menu_Generator::quit(unsigned int first_shown_index,
                     unsigned int selected_index)
{
    // ??
}

void
Menu_Generator::update_screen_shot(unsigned int selected_index,
                                   int brightness)
{
#if 0
        if ((p_selected_index % 7) == 0)
            p_screen_shot->load_bitmap("shot.pcx");
        else if ((p_selected_index % 7) == 1)
            p_screen_shot->set_to_black();
        else if ((p_selected_index % 7) == 2)
            p_screen_shot->load_bitmap("pacman.pcx");
        else if ((p_selected_index % 7) == 2)
            p_screen_shot->load_bitmap("robotron.pcx");
        else if ((p_selected_index % 7) == 3)
            p_screen_shot->load_bitmap("galaga.pcx");
        else if ((p_selected_index % 7) == 4)
            p_screen_shot->load_bitmap("tmnt.pcx");
        else if ((p_selected_index % 7) == 5)
            p_screen_shot->load_bitmap("ecco2.pcx");
        else if ((p_selected_index % 7) == 6)
            p_screen_shot->load_bitmap("pengo.pcx");
#endif
}

bool
Menu_Generator::displaying_screen_shot()
{
    return true;
}

void
Menu_Generator::set_text(char * text, const char * title, int num)
{
    sprintf(text, "%s selection %0*d", title, ((num % 15) + 1), (num + 1));
    return;
}


void
Menu_Generator::setup_menu()
{
    Menu_Description *  description;

    description = &p_menus[p_current_menu];

    p_menu->set_menu(description->title, description->sub_title,
                     description->games, this, description->first_shown_index,
                     description->selected_index);
    return;
}
#endif
