#include "sshot.h"
#include "logfile.h"
#if (USE_JPG == 1)
#include "jpgalleg.h"
#endif

using namespace GameLauncher;
using LogFile::gl_log;

#if (USE_JPG == 1)
extern "C"
{
    static void jpeg_progress_monitor(void * data, int completed, int total);
}

static void
jpeg_progress_monitor(void * data, int completed, int total)
{
    Engine *    engine;
    engine = (Engine *) data;
    engine->poll_background_task();
}
#endif

Screen_Shot::Screen_Shot(Engine * engine)
{
    bool        try_vram;

    p_x         = 0;
    p_y         = 0;
    p_width     = SCREEN_W;
    p_height    = SCREEN_H;


#if (USE_JPG == 1)
    jpg_register_progress_monitor(engine, jpeg_progress_monitor);
#endif

    try_vram = ((engine->use_vram()) &&
                (engine->animation_method() == Engine::PAGE_FLIPPING) &&
                ((gfx_capabilities & GFX_HW_VRAM_BLIT) != 0));

    p_screen_shot               = 0;
    p_unmodified_screen_shot    = 0;

    if (try_vram)
    {
        p_screen_shot = create_video_bitmap(p_width, p_height);
        if (p_screen_shot != 0)
            gl_log->info("VRAM bitmap for screen shot");
    }

    // Fall back to memory bitmaps
    if (p_screen_shot == 0)
    {
        if (try_vram)
            gl_log->info("No VRAM: memory bitmap for screen shot");
        else
            gl_log->info("No HW blit: memory bitmap for screen shot");
        p_screen_shot = create_bitmap(p_width, p_height);
    }

    // Nothin' worked...
    if (p_screen_shot == 0)
        return;

    clear(p_screen_shot);

    p_null_listener     = new Null_Screen_Shot_Listener();
    p_listener          = p_null_listener;

    p_toggle_key        =
        engine->get_keypress_factory()->create_keypress_watcher(0);
    p_toggle_key->set_one_time();

    p_delay_timer       =
        engine->get_timer_factory()->create_timer();

    p_bitmap_file       = "";
    p_new_bitmap_file   = "";
    p_brightness_percentage     = 0;
    p_use_bitmap        = false;
    
    return;
}

Screen_Shot::~Screen_Shot()
{
    if (p_screen_shot != 0)
    {
        destroy_bitmap(p_screen_shot);
        p_screen_shot = 0;

        delete p_null_listener;
        delete p_toggle_key;
        delete p_delay_timer;
    }

    return;
}

void
Screen_Shot::set_bitmap(const string & bitmap_file, fix brightness_percentage,
                        int delay = 0)
{
    p_new_bitmap_file           = bitmap_file;
    p_brightness_percentage     = brightness_percentage;

    if (delay == 0)
    {
        p_delay_timer->disable();
        update_bitmap();
        p_state = SHOWING;
    }
    else
    {
        p_delay_timer->set(delay);
        p_state = WAITING;
    }
}

void
Screen_Shot::cancel_set_bitmap()
{
    if (p_state == WAITING)
    {
        // Reset back to current showing file
        p_new_bitmap_file       = p_bitmap_file;
        p_delay_timer->disable();
        p_state = SHOWING;
    }
}


void
Screen_Shot::update_bitmap()
{
    PALETTE     p;
    bool        need_to_load_bitmap;

    need_to_load_bitmap = false;
    if (p_new_bitmap_file != p_bitmap_file)
    {
        p_bitmap_file   = p_new_bitmap_file;
        need_to_load_bitmap     = true;
    }

    // If the screen shot will be black, just set a flag and return.
    if ((p_bitmap_file == "") || (p_brightness_percentage == 0))
    {
        p_use_bitmap = false;
        return;
    }

    // Load bitmap, if necessary.
    if (need_to_load_bitmap)
    {
        p_bitmap_file = p_new_bitmap_file;
        if (p_unmodified_screen_shot != 0)
            destroy_bitmap(p_unmodified_screen_shot);

        p_unmodified_screen_shot =
            ::load_bitmap((char *) p_bitmap_file.c_str(), p);
        if (p_unmodified_screen_shot == 0)
        {
            return;
        }
    }

    BITMAP *    shaded_bmp;
    BITMAP *    tmp_bmp;
    int         int_scale_w, int_scale_h, int_scale;
    double      scale_w, scale_h, scale;
    int         nx, ny, nw, nh;

    if (p_unmodified_screen_shot == 0)
        return;

    int_scale = 0;
    if (p_prefer_integer_scaling)
    {
        int_scale_w = p_screen_shot->w / p_unmodified_screen_shot->w;
        int_scale_h = p_screen_shot->h / p_unmodified_screen_shot->h;
        int_scale   = MIN(int_scale_w, int_scale_h);
    }

    if (int_scale != 0)
    {
        nw = p_unmodified_screen_shot->w * int_scale;
        nh = p_unmodified_screen_shot->h * int_scale;
    }
    else
    {
        // Use floating point scale if integer scaling failed or
        // integer scaling is not preferred.
        scale_w = (double(p_screen_shot->w) /
                   double(p_unmodified_screen_shot->w));
        scale_h = (double(p_screen_shot->h) /
                   double(p_unmodified_screen_shot->h));
        scale   = MIN(scale_w, scale_h);
        nw      = int( double(p_unmodified_screen_shot->w) * scale);
        nh      = int( double(p_unmodified_screen_shot->h) * scale);
    }

    nx = (p_screen_shot->w - nw) / 2;
    ny = (p_screen_shot->h - nh) / 2;

    shaded_bmp = 0;

    if (p_brightness_percentage != 100)
    {
        int     alpha;

        // Scale percentage to value between 0 and 255
        alpha = (p_brightness_percentage * fix(255)) / fix(100);

        shaded_bmp = create_bitmap(p_unmodified_screen_shot->w,
                                   p_unmodified_screen_shot->h);
        if (shaded_bmp == 0)
        {
            return;
        }
        clear(shaded_bmp);
        set_trans_blender(0, 0, 0, alpha);
        draw_trans_sprite(shaded_bmp, p_unmodified_screen_shot, 0, 0);
        tmp_bmp = shaded_bmp;
    }
    else
        tmp_bmp = p_unmodified_screen_shot;

    clear(p_screen_shot);
    stretch_blit(tmp_bmp, p_screen_shot, 0, 0, tmp_bmp->w, tmp_bmp->h,
                 nx, ny, nw, nh);

    if (shaded_bmp != 0)
        destroy_bitmap(shaded_bmp);

    p_use_bitmap = true;
    return;
}

void
Screen_Shot::draw(BITMAP * frame)
{
    if (p_use_bitmap)
    {
        // clear(frame);
        blit(p_screen_shot, frame, 0, 0, p_x, p_y, p_screen_shot->w,
             p_screen_shot->h);
    }
    else
    {
        clear(frame);
    }

    return;
}

bool
Screen_Shot::update()
{
    bool        updated;

    updated     = false;
    if (p_toggle_key->is_pressed())
        p_listener->toggle_screen_shot();

    if ((p_state == WAITING) && (p_delay_timer->expired()))
    {
        update_bitmap();
        p_state = SHOWING;
        updated = true;
    }

    return updated;
}

void
Screen_Shot::set_listener(Screen_Shot_Listener * listener)
{
    if (listener != 0)
        p_listener      = listener;
    else
        p_listener      = p_null_listener;
}

void
Screen_Shot::set_toggle_key(int scancode)
{
    p_toggle_key->set_scancode(scancode);
}

void
Screen_Shot::set_prefer_integer_scaling(bool prefer_integer_scaling)
{
    p_prefer_integer_scaling = prefer_integer_scaling;
}

void
Null_Screen_Shot_Listener::toggle_screen_shot()
{
}
